import os
import sys
import threading

from pydantic import ValidationError

from xconn._client import helpers
from xconn.types import WebsocketConfig
from xconn._client.helpers import connect
from xconn._client.watcher import start_file_watcher
from xconn._client.types import ClientConfig, CommandArgs, ConfigSource


def handle_start(command_args: CommandArgs):
    if command_args.config_source == ConfigSource.cli:
        try:
            config = ClientConfig(
                url=command_args.url,
                realm=command_args.realm,
                authid=command_args.authid,
                secret=command_args.secret,
                ticket=command_args.ticket,
                private_key=command_args.private_key,
            )
        except ValidationError as e:
            print("Invalid input:")
            for error in e.errors():
                loc = " -> ".join(map(str, error["loc"]))
                msg = error["msg"]
                print(f" - {loc}: {msg}")

            exit(1)

        config.websocket_config = WebsocketConfig(
            command_args.open_timeout, command_args.ping_interval, command_args.ping_timeout
        )
    elif command_args.config_source == ConfigSource.env:
        config = helpers.load_config_from_env(command_args)
        # Now override config if something was provided explicitly from command line
        config = helpers.update_config_from_cli(config, command_args)
    elif command_args.config_source == ConfigSource.file:
        config = helpers.load_config_from_file(command_args)
        # Now override config if something was provided explicitly from command line
        config = helpers.update_config_from_cli(config, command_args)
    else:
        raise RuntimeError(f"Unknown config_source: {command_args.config_source}")

    helpers.validate_auth_inputs(config)
    config.authmethod = helpers.select_authmethod(config)

    if command_args.reload:

        def restart():
            # Restart process due to file changes
            os.execv(sys.executable, [sys.executable] + sys.argv)

        threading.Thread(target=start_file_watcher, args=(restart,), daemon=True).start()

    connect(command_args.app, config, start_router=command_args.start_router, directory=command_args.directory)



import asyncio
import os
import socket
from asyncio import StreamReader, StreamWriter, Future
from concurrent.futures import Future as ConcurrentFuture
from dataclasses import dataclass
import time
from typing import Sequence
from urllib.parse import urlparse

from wampproto.transports.rawsocket import (
    Handshake,
    MessageHeader,
    DEFAULT_MAX_MSG_SIZE,
    SERIALIZER_TYPE_CBOR,
    MSG_TYPE_WAMP,
    MSG_TYPE_PING,
    MSG_TYPE_PONG,
)
from websockets import State, Subprotocol
from websockets.sync.client import connect, unix_connect
from websockets.sync.connection import Connection
from websockets.asyncio.client import connect as async_connect, unix_connect as async_unix_connect
from websockets.asyncio.client import ClientConnection

from xconn.types import IAsyncTransport, ITransport, WebsocketConfig

# Applies to handshake and message itself.
RAW_SOCKET_HEADER_LENGTH = 4


@dataclass
class PendingPing:
    future: Future[float] | ConcurrentFuture[float]
    created_at: float


def create_ping():
    payload = os.urandom(16)
    ping_header = MessageHeader(MSG_TYPE_PING, len(payload))
    created_at = time.time() * 1000

    return payload, ping_header, created_at

def _recv_exactly(sock, n: int) -> bytes:
    """Receive exactly n bytes from a socket or raise if connection breaks."""
    chunks = []
    received = 0
    while received < n:
        chunk = sock.recv(n - received)
        if not chunk:
            raise ConnectionError("Socket connection broken")

        chunks.append(chunk)
        received += len(chunk)

    return b"".join(chunks)


class RawSocketTransport(ITransport):
    def __init__(self, sock: socket.socket):
        super().__init__()
        self._sock = sock

        self._pending_pings: dict[bytes, PendingPing] = {}

    @staticmethod
    def connect(
        uri: str, protocol: int = SERIALIZER_TYPE_CBOR, max_msg_size: int = DEFAULT_MAX_MSG_SIZE
    ) -> "RawSocketTransport":
        parsed = urlparse(uri)

        if parsed.scheme == "rs" or parsed.scheme == "rss" or parsed.scheme == "tcp" or parsed.scheme == "tcps":
            sock = socket.create_connection((parsed.hostname, parsed.port))
        elif parsed.scheme == "unix" or parsed.scheme == "unix+rs":
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(parsed.path)
        else:
            raise RuntimeError(f"Unsupported scheme {parsed.scheme}")

        hs_request = Handshake(protocol, max_msg_size)

        sock.sendall(hs_request.to_bytes())

        hs_response_bytes = _recv_exactly(sock, RAW_SOCKET_HEADER_LENGTH)
        hs_response = Handshake.from_bytes(hs_response_bytes)

        if hs_request.protocol != hs_response.protocol:
            raise ValueError("Handshake protocol mismatch.")

        return RawSocketTransport(sock)

    def read(self) -> str | bytes:
        msg_header_bytes = _recv_exactly(self._sock, RAW_SOCKET_HEADER_LENGTH)
        msg_header = MessageHeader.from_bytes(msg_header_bytes)

        if msg_header.kind == MSG_TYPE_WAMP:
            return _recv_exactly(self._sock, msg_header.length)
        elif msg_header.kind == MSG_TYPE_PING:
            ping_payload = _recv_exactly(self._sock, msg_header.length)
            pong = MessageHeader(MSG_TYPE_PONG, msg_header.length)
            self._sock.sendall(pong.to_bytes())
            self._sock.sendall(ping_payload)

            return self.read()
        elif msg_header.kind == MSG_TYPE_PONG:
            pong_payload = _recv_exactly(self._sock, msg_header.length)
            pending_ping = self._pending_pings.pop(pong_payload, None)
            if pending_ping is not None:
                received_at = time.time() * 1000
                pending_ping.future.set_result(received_at - pending_ping.created_at)

            return self.read()
        else:
            raise ValueError(f"Unsupported message type {msg_header.kind}")

    def write(self, data: str | bytes):
        msg_header = MessageHeader(MSG_TYPE_WAMP, len(data))

        self._sock.sendall(msg_header.to_bytes())

        if isinstance(data, str):
            self._sock.sendall(data.encode())
        else:
            self._sock.sendall(data)

    def close(self):
        self._sock.close()

    def is_connected(self) -> bool:
        try:
            self._sock.send(b"")  # Send zero bytes
            return True
        except (BrokenPipeError, ConnectionResetError, OSError):
            return False

    def ping(self, timeout: int = 10) -> float:
        f: ConcurrentFuture[int] = ConcurrentFuture()
        payload, ping_header, created_at = create_ping()
        self._pending_pings[payload] = PendingPing(f, created_at)

        self._sock.sendall(ping_header.to_bytes())
        self._sock.sendall(payload)

        return f.result(timeout)


class AsyncRawSocketTransport(IAsyncTransport):
    def __init__(self, reader: StreamReader, writer: StreamWriter):
        super().__init__()
        self._reader = reader
        self._writer = writer

        self._pending_pings: dict[bytes, PendingPing] = {}

    @staticmethod
    async def connect(
        uri: str, protocol: int = SERIALIZER_TYPE_CBOR, max_msg_size: int = DEFAULT_MAX_MSG_SIZE
    ) -> "AsyncRawSocketTransport":
        parsed = urlparse(uri)

        if parsed.scheme == "rs" or parsed.scheme == "rss" or parsed.scheme == "tcp" or parsed.scheme == "tcps":
            reader, writer = await asyncio.open_connection(parsed.hostname, parsed.port)
        elif parsed.scheme == "unix" or parsed.scheme == "unix+rs":
            reader, writer = await asyncio.open_unix_connection(parsed.path)
        else:
            raise RuntimeError(f"Unsupported scheme {parsed.scheme}")

        hs_request = Handshake(protocol, max_msg_size)

        writer.write(hs_request.to_bytes())
        await writer.drain()

        hs_response_bytes = await reader.readexactly(RAW_SOCKET_HEADER_LENGTH)
        hs_response = Handshake.from_bytes(hs_response_bytes)

        if hs_request.protocol != hs_response.protocol:
            raise ValueError("Handshake protocol mismatch.")

        return AsyncRawSocketTransport(reader, writer)

    async def read(self) -> str | bytes:
        msg_header_bytes = await self._reader.readexactly(RAW_SOCKET_HEADER_LENGTH)
        msg_header = MessageHeader.from_bytes(msg_header_bytes)

        if msg_header.kind == MSG_TYPE_WAMP:
            return await self._reader.readexactly(msg_header.length)
        elif msg_header.kind == MSG_TYPE_PING:
            ping_payload = await self._reader.readexactly(msg_header.length)
            pong = MessageHeader(MSG_TYPE_PONG, msg_header.length)
            self._writer.write(pong.to_bytes())
            await self._writer.drain()
            self._writer.write(ping_payload)
            await self._writer.drain()

            return await self.read()
        elif msg_header.kind == MSG_TYPE_PONG:
            pong_payload = await self._reader.readexactly(msg_header.length)
            pending_ping = self._pending_pings.pop(pong_payload, None)
            if pending_ping is not None:
                received_at = time.time() * 1000
                pending_ping.future.set_result(received_at - pending_ping.created_at)

            return await self.read()
        else:
            raise ValueError(f"Unsupported message type {msg_header.kind}")

    async def write(self, data: str | bytes):
        msg_header = MessageHeader(MSG_TYPE_WAMP, len(data))

        self._writer.write(msg_header.to_bytes())
        await self._writer.drain()

        if isinstance(data, str):
            self._writer.write(data.encode())
        else:
            self._writer.write(data)

        await self._writer.drain()

    async def close(self):
        self._writer.close()

    async def is_connected(self) -> bool:
        try:
            self._writer.write(b"")  # Send zero bytes
            await self._writer.drain()
            return True
        except (BrokenPipeError, ConnectionResetError, OSError):
            return False

    async def ping(self, timeout: int = 10) -> float:
        f: Future[int] = Future()
        payload, ping_header, created_at = create_ping()
        self._pending_pings[payload] = PendingPing(f, created_at)

        self._writer.write(ping_header.to_bytes())
        await self._writer.drain()

        self._writer.write(payload)
        await self._writer.drain()

        return await asyncio.wait_for(f, timeout)


class WebSocketTransport(ITransport):
    def __init__(self, websocket: Connection):
        super().__init__()
        self._websocket = websocket

    @staticmethod
    def connect(uri: str, subprotocols: Sequence[Subprotocol], config: WebsocketConfig) -> "WebSocketTransport":
        parsed_url = urlparse(uri)
        if parsed_url.scheme == "unix+ws":
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(parsed_url.path)
            ws = unix_connect(
                parsed_url.path,
                subprotocols=subprotocols,
                open_timeout=config.open_timeout,
                ping_interval=config.ping_interval,
                ping_timeout=config.ping_timeout,
                close_timeout=config.close_timeout,
            )
        else:
            ws = connect(
                uri,
                subprotocols=subprotocols,
                open_timeout=config.open_timeout,
                ping_interval=config.ping_interval,
                ping_timeout=config.ping_timeout,
                close_timeout=config.close_timeout,
            )

        return WebSocketTransport(ws)

    def read(self) -> str | bytes:
        return self._websocket.recv()

    def write(self, data: str | bytes):
        self._websocket.send(data)

    def close(self):
        self._websocket.close()

    def is_connected(self) -> bool:
        return self._websocket.state == State.OPEN

    def ping(self, timeout: int = 10) -> float:
        payload, _, created_at = create_ping()

        event = self._websocket.ping(payload)
        event.wait(timeout)
        received_at = time.time() * 1000
        return received_at - created_at


class AsyncWebSocketTransport(IAsyncTransport):
    def __init__(self, websocket: ClientConnection):
        super().__init__()
        self._websocket = websocket

    @staticmethod
    async def connect(
        uri: str, subprotocols: Sequence[Subprotocol], config: WebsocketConfig
    ) -> "AsyncWebSocketTransport":
        parsed_url = urlparse(uri)
        if parsed_url.scheme == "unix+ws":
            ws = await async_unix_connect(
                parsed_url.path,
                subprotocols=subprotocols,
                open_timeout=config.open_timeout,
                ping_interval=config.ping_interval,
                ping_timeout=config.ping_timeout,
                close_timeout=config.close_timeout,
            )
        else:
            ws = await async_connect(
                uri,
                subprotocols=subprotocols,
                open_timeout=config.open_timeout,
                ping_interval=config.ping_interval,
                ping_timeout=config.ping_timeout,
                close_timeout=config.close_timeout,
            )

        return AsyncWebSocketTransport(ws)

    async def read(self) -> str | bytes:
        return await self._websocket.recv()

    async def write(self, data: str | bytes):
        await self._websocket.send(data)

    async def close(self):
        await self._websocket.close()

    async def is_connected(self) -> bool:
        return self._websocket.state == State.OPEN

    async def ping(self, timeout: int = 10) -> float:
        payload, _, created_at = create_ping()

        awaitable = await self._websocket.ping(payload)
        await asyncio.wait_for(awaitable, timeout)
        received_at = time.time() * 1000
        return received_at - created_at



from wampproto import joiner, serializers, auth
from wampproto.joiner import Joiner

from xconn import types, helpers
from xconn.transports import WebSocketTransport, AsyncWebSocketTransport, RawSocketTransport, AsyncRawSocketTransport


class WebsocketsJoiner:
    def __init__(
        self,
        authenticator: auth.IClientAuthenticator = None,
        serializer: serializers.Serializer = serializers.JSONSerializer(),
        ws_config: types.WebsocketConfig = types.WebsocketConfig(),
    ):
        self._authenticator = authenticator
        self._serializer = serializer
        self._ws_config = ws_config

    def join(self, uri: str, realm: str) -> types.BaseSession:
        transport = WebSocketTransport.connect(
            uri,
            subprotocols=[helpers.get_ws_subprotocol(serializer=self._serializer)],
            config=self._ws_config,
        )

        j: Joiner = joiner.Joiner(realm, serializer=self._serializer, authenticator=self._authenticator)
        transport.write(j.send_hello())

        while True:
            data = transport.read()
            to_send = j.receive(data)
            if to_send is None:
                return types.BaseSession(transport, j.get_session_details(), self._serializer)

            transport.write(to_send)


class AsyncWebsocketsJoiner:
    def __init__(
        self,
        authenticator: auth.IClientAuthenticator = None,
        serializer: serializers.Serializer = serializers.JSONSerializer(),
        ws_config: types.WebsocketConfig = types.WebsocketConfig(),
    ):
        self._ws_config = ws_config
        self._authenticator = authenticator
        self._serializer = serializer

    async def join(self, uri: str, realm: str) -> types.AsyncBaseSession:
        transport = await AsyncWebSocketTransport.connect(
            uri,
            subprotocols=[helpers.get_ws_subprotocol(serializer=self._serializer)],
            config=self._ws_config,
        )

        j: Joiner = joiner.Joiner(realm, serializer=self._serializer, authenticator=self._authenticator)
        await transport.write(j.send_hello())

        while True:
            data = await transport.read()
            to_send = j.receive(data)
            if to_send is None:
                return types.AsyncBaseSession(transport, j.get_session_details(), self._serializer)

            await transport.write(to_send)


class RawSocketJoiner:
    def __init__(
        self,
        authenticator: auth.IClientAuthenticator = None,
        serializer: serializers.Serializer = serializers.JSONSerializer(),
    ):
        self._authenticator = authenticator
        self._serializer = serializer

    def join(self, uri: str, realm: str) -> types.BaseSession:
        transport = RawSocketTransport.connect(uri, helpers.get_rs_protocol(self._serializer))

        j: Joiner = joiner.Joiner(realm, serializer=self._serializer, authenticator=self._authenticator)
        transport.write(j.send_hello())

        while True:
            data = transport.read()
            to_send = j.receive(data)
            if to_send is None:
                return types.BaseSession(transport, j.get_session_details(), self._serializer)

            transport.write(to_send)


class AsyncRawSocketJoiner:
    def __init__(
        self,
        authenticator: auth.IClientAuthenticator = None,
        serializer: serializers.Serializer = serializers.JSONSerializer(),
    ):
        self._authenticator = authenticator
        self._serializer = serializer

    async def join(self, uri: str, realm: str) -> types.AsyncBaseSession:
        transport = await AsyncRawSocketTransport.connect(uri, helpers.get_rs_protocol(self._serializer))
        j: Joiner = joiner.Joiner(realm, serializer=self._serializer, authenticator=self._authenticator)
        await transport.write(j.send_hello())

        while True:
            data = await transport.read()
            to_send = j.receive(data)
            if to_send is None:
                return types.AsyncBaseSession(transport, j.get_session_details(), self._serializer)

            await transport.write(to_send)



import json
import inspect
from typing import Any, List as TypedList, Dict as TypingDict


class Value:
    def __init__(self, data: Any):
        self.data = data

    def raw(self) -> Any:
        return self.data

    def string(self) -> str:
        if isinstance(self.data, str):
            return self.data

        raise TypeError(f"value is not data string, got {type(self.data).__name__}")

    def string_or(self, default: str) -> str:
        try:
            return self.string()
        except TypeError:
            return default

    def bool_(self) -> bool:
        if isinstance(self.data, bool):
            return self.data

        raise TypeError(f"value is not data bool, got {type(self.data).__name__}")

    def bool_or(self, default: bool) -> bool:
        try:
            return self.bool_()
        except TypeError:
            return default

    def float_(self) -> float:
        if isinstance(self.data, (int, float)):
            return float(self.data)

        raise TypeError(f"value cannot be converted to float, got {type(self.data).__name__}")

    def float_or(self, default: float) -> float:
        try:
            return self.float_()
        except TypeError:
            return default

    def bytes_(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        if isinstance(self.data, str):
            return self.data.encode()
        raise TypeError(f"value cannot be converted to bytes, got {type(self.data).__name__}")

    def bytes_or(self, default: bytes) -> bytes:
        try:
            return self.bytes_()
        except TypeError:
            return default

    def decode(self, out_type: Any) -> Any:
        try:
            serialized = json.dumps(self.data)
            return json.loads(serialized, object_hook=out_type)
        except Exception as e:
            raise ValueError(f"failed to decode value: {e}")

    def list(self) -> TypedList[Any]:
        if isinstance(self.data, list):
            return self.data

        raise TypeError(f"value is not data list, got {type(self.data).__name__}")

    def list_or(self, default: TypedList[Any]) -> TypedList[Any]:
        try:
            return self.list()
        except TypeError:
            return default


class List:
    def __init__(self, values: TypedList[Any]):
        self.values: TypedList[Value] = [Value(v) for v in values]

    def __len__(self) -> int:
        return len(self.values)

    def get(self, i: int) -> Value:
        if 0 <= i < len(self.values):
            return self.values[i]

        raise IndexError(f"index {i} out of range [0, {len(self.values)})")

    def get_or(self, i: int, default: Any) -> Value:
        try:
            return self.get(i)
        except IndexError:
            return Value(default)

    def string(self, i: int) -> str:
        return self.get(i).string()

    def string_or(self, i: int, default: str) -> str:
        return self.get_or(i, default).string_or(default)

    def bool_(self, i: int) -> bool:
        return self.get(i).bool_()

    def bool_or(self, i: int, default: bool) -> bool:
        return self.get_or(i, default).bool_or(default)

    def float_(self, i: int) -> float:
        return self.get(i).float_()

    def float_or(self, i: int, default: float) -> float:
        return self.get_or(i, default).float_or(default)

    def bytes_(self, i: int) -> bytes:
        return self.get(i).bytes_()

    def bytes_or(self, i: int, default: bytes) -> bytes:
        return self.get_or(i, default).bytes_or(default)

    def decode(self, out_type: Any) -> Any:
        raw_values = [v.raw() for v in self.values]

        if not inspect.isclass(out_type):
            raise TypeError("out_type must be a class")

        try:
            serialized = json.dumps(raw_values)
            return json.loads(serialized, object_hook=lambda d: out_type(**d))
        except Exception as e:
            raise ValueError(f"failed to decode list: {e}")

    def list(self, i: int) -> TypedList[Any]:
        return self.get(i).list()

    def list_or(self, i: int, default: TypedList[Any]) -> TypedList[Any]:
        return self.get_or(i, default).list_or(default)

    def raw(self) -> TypedList[Any]:
        return [v.raw() for v in self.values]


class Dict:
    def __init__(self, values: TypingDict[str, Any]):
        self.values: TypingDict[str, Value] = {k: Value(v) for k, v in values.items()}

    def __len__(self) -> int:
        return len(self.values)

    def has(self, key: str) -> bool:
        return key in self.values

    def get(self, key: str) -> Value:
        if key in self.values:
            return self.values[key]

        raise KeyError(f"key '{key}' not found")

    def get_or(self, key: str, default: Any) -> Value:
        return self.values.get(key, Value(default))

    def string(self, key: str) -> str:
        return self.get(key).string()

    def string_or(self, key: str, default: str) -> str:
        return self.get_or(key, default).string_or(default)

    def bool_(self, key: str) -> bool:
        return self.get(key).bool_()

    def bool_or(self, key: str, default: bool) -> bool:
        return self.get_or(key, default).bool_or(default)

    def float_(self, key: str) -> float:
        return self.get(key).float_()

    def float_or(self, key: str, default: float) -> float:
        return self.get_or(key, default).float_or(default)

    def bytes_(self, key: str) -> bytes:
        return self.get(key).bytes_()

    def bytes_or(self, key: str, default: bytes) -> bytes:
        return self.get_or(key, default).bytes_or(default)

    def raw(self) -> TypingDict[str, Any]:
        return {k: v.raw() for k, v in self.values.items()}

    def decode(self, out_type: Any) -> Any:
        try:
            serialized = json.dumps(self.raw())
            if not inspect.isclass(out_type):
                raise TypeError("out_type must be a class")

            return json.loads(serialized, object_hook=lambda d: out_type(**d))
        except Exception as e:
            raise ValueError(f"failed to decode dict into structure: {e}")



from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler


class FileChangeHandler(FileSystemEventHandler):
    def __init__(self, on_change):
        super().__init__()
        self.on_change = on_change

    def on_any_event(self, event):
        if event.src_path.endswith(".py"):
            print(f"Detected change in {event.src_path}")
            self.on_change()


def start_file_watcher(on_change, directory: str = "."):
    observer = Observer()
    handler = FileChangeHandler(on_change)
    observer.schedule(handler, path=directory, recursive=True)
    observer.start()

    return observer



from xconn.app import App, Component, register, subscribe
from xconn.client import Client
from xconn.async_client import AsyncClient
from xconn.router import Router
from xconn.server import Server
from xconn.utils import run
from xconn._client.helpers import connect

from wampproto.auth import (
    AnonymousAuthenticator,
    TicketAuthenticator,
    WAMPCRAAuthenticator,
    CryptoSignAuthenticator,
)

from wampproto.serializers import (
    JSONSerializer,
    MsgPackSerializer,
    CBORSerializer,
)

__all__ = [
    "App",
    "Component",
    "register",
    "subscribe",
    "Client",
    "AsyncClient",
    "Router",
    "Server",
    # export authenticators
    "AnonymousAuthenticator",
    "TicketAuthenticator",
    "WAMPCRAAuthenticator",
    "CryptoSignAuthenticator",
    # export serializers
    "JSONSerializer",
    "MsgPackSerializer",
    "CBORSerializer",
    # runner for async client
    "run",
    # xcorn connect function
    "connect",
]

_capnp_import_error = None
try:
    from xconn.helpers import CapnProtoSerializer

    __all__.extend(["CapnProtoSerializer"])
except ImportError as e:
    _capnp_import_error = e

    class CapnProtoSerializer:
        def __init__(self, *args, **kwargs):
            raise ImportError(
                "Cap'n Proto serializer support is not installed.\nInstall it with:\n  uv pip install xconn[capnproto]"
            ) from _capnp_import_error



import argparse

from xconn._client.cli import handle_start, CommandArgs
from xconn._client.types import ConfigSource


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("APP", type=str)
    parser.add_argument("--url", type=str)
    parser.add_argument("--realm", type=str)
    parser.add_argument("--authid", type=str)
    parser.add_argument("--secret", type=str)
    parser.add_argument("--ticket", type=str)
    parser.add_argument("--private-key", type=str)

    # transport options
    parser.add_argument("--open-timeout", type=int, default=10)
    parser.add_argument("--ping-interval", type=int, default=20)
    parser.add_argument("--ping-timeout", type=int, default=20)

    # misc
    parser.add_argument("--directory", type=str, default=".")
    parser.add_argument("--start-router", action="store_true", default=False)
    parser.add_argument("--reload", action="store_true", default=False)
    parser.add_argument(
        "--config-source", type=str, choices=[e.value for e in ConfigSource], default=ConfigSource.cli.value
    )
    parser.add_argument("--config-file", type=str)
    args = parser.parse_args()

    handle_start(CommandArgs(**vars(args)))



from __future__ import annotations

import asyncio
import inspect
from dataclasses import dataclass
from asyncio import Future, get_event_loop
from typing import Callable, Union, Awaitable, Any

from wampproto import messages, idgen, session

from xconn import types, uris as xconn_uris, exception
from xconn.exception import ApplicationError
from xconn.helpers import exception_from_error


@dataclass
class RegisterRequest:
    future: Future[Registration]
    endpoint: Callable | Callable[[types.Invocation], Awaitable[types.Result]]


class Registration:
    def __init__(self, registration_id: int, session: AsyncSession):
        self.registration_id = registration_id
        self._session = session

    async def unregister(self) -> None:
        if not await self._session._base_session.transport.is_connected():
            raise Exception("cannot unregister procedure: session not established")

        unregister = messages.Unregister(messages.UnregisterFields(self._session._idgen.next(), self.registration_id))
        data = self._session._session.send_message(unregister)

        f: Future = Future()
        self._session._unregister_requests[unregister.request_id] = types.UnregisterRequest(f, self.registration_id)
        await self._session._base_session.send(data)

        return await f


@dataclass
class SubscribeRequest:
    future: Future[Subscription]
    endpoint: Callable[[types.Event], Awaitable[None]]


class Subscription:
    def __init__(self, subscription_id: int, session: AsyncSession):
        self.subscription_id = subscription_id
        self._session = session

    async def unsubscribe(self) -> None:
        if not await self._session._base_session.transport.is_connected():
            raise Exception("cannot unsubscribe topic: session not established")

        unsubscribe = messages.Unsubscribe(
            messages.UnsubscribeFields(self._session._idgen.next(), self.subscription_id)
        )
        data = self._session._session.send_message(unsubscribe)

        f: Future = Future()
        self._session._unsubscribe_requests[unsubscribe.request_id] = types.UnsubscribeRequest(f, self.subscription_id)
        await self._session._base_session.send(data)

        return await f


class AsyncSession:
    def __init__(self, base_session: types.IAsyncBaseSession):
        # RPC data structures
        self._call_requests: dict[int, Future[types.Result]] = {}
        self._register_requests: dict[int, RegisterRequest] = {}
        self._registrations: dict[
            int,
            Union[Callable[[types.Invocation], types.Result], Callable[[types.Invocation], Awaitable[types.Result]]],
        ] = {}
        self._unregister_requests: dict[int, types.UnregisterRequest] = {}

        # PubSub data structures
        self._publish_requests: dict[int, Future[None]] = {}
        self._subscribe_requests: dict[int, SubscribeRequest] = {}
        self._subscriptions: dict[int, Callable[[types.Event], Awaitable[None]]] = {}
        self._unsubscribe_requests: dict[int, types.UnsubscribeRequest] = {}

        self._goodbye_request = Future()

        # ID generator
        self._idgen = idgen.SessionScopeIDGenerator()

        self._base_session = base_session

        # initialize the sans-io wamp session
        self._session = session.WAMPSession(base_session.serializer)

        self._disconnect_callback: list[Callable[[], Awaitable[None]] | None] = []

        loop = get_event_loop()
        self.wait_task = loop.create_task(self._wait())

    async def _wait(self):
        while await self._base_session.transport.is_connected():
            try:
                data = await self._base_session.receive()
            except Exception as e:
                print(e)
                break

            await self._process_incoming_message(self._session.receive(data))

        for callback in self._disconnect_callback:
            await callback()

    async def _process_incoming_message(self, msg: messages.Message):
        if isinstance(msg, messages.Registered):
            request = self._register_requests.pop(msg.request_id)
            self._registrations[msg.registration_id] = request.endpoint
            request.future.set_result(Registration(msg.registration_id, self))
        elif isinstance(msg, messages.Unregistered):
            request = self._unregister_requests.pop(msg.request_id)
            del self._registrations[request.registration_id]
            request.future.set_result(None)
        elif isinstance(msg, messages.Result):
            request = self._call_requests.pop(msg.request_id)
            request.set_result(types.Result(msg.args, msg.kwargs, msg.details))
        elif isinstance(msg, messages.Invocation):
            try:
                endpoint = self._registrations[msg.registration_id]
                result = await endpoint(types.Invocation(msg.args, msg.kwargs, msg.details))

                if result is None:
                    data = self._session.send_message(messages.Yield(messages.YieldFields(msg.request_id)))
                elif isinstance(result, types.Result):
                    data = self._session.send_message(
                        messages.Yield(messages.YieldFields(msg.request_id, result.args, result.kwargs, result.details))
                    )
                else:
                    message = "Endpoint returned invalid result type. Expected types.Result or None, got: " + str(
                        type(result)
                    )
                    msg_to_send = messages.Error(
                        messages.ErrorFields(msg.TYPE, msg.request_id, xconn_uris.ERROR_INTERNAL_ERROR, [message])
                    )
                    data = self._session.send_message(msg_to_send)

                await self._base_session.send(data)
            except ApplicationError as e:
                msg_to_send = messages.Error(messages.ErrorFields(msg.TYPE, msg.request_id, e.message, e.args))
                data = self._session.send_message(msg_to_send)
                await self._base_session.send(data)
            except Exception as e:
                msg_to_send = messages.Error(
                    messages.ErrorFields(msg.TYPE, msg.request_id, xconn_uris.ERROR_RUNTIME_ERROR, [e.__str__()])
                )
                data = self._session.send_message(msg_to_send)
                await self._base_session.send(data)
        elif isinstance(msg, messages.Subscribed):
            request = self._subscribe_requests.pop(msg.request_id)
            self._subscriptions[msg.subscription_id] = request.endpoint
            request.future.set_result(Subscription(msg.subscription_id, self))
        elif isinstance(msg, messages.Unsubscribed):
            request = self._unsubscribe_requests.pop(msg.request_id)
            del self._subscriptions[request.subscription_id]
            request.future.set_result(None)
        elif isinstance(msg, messages.Published):
            request = self._publish_requests.pop(msg.request_id)
            request.set_result(None)
        elif isinstance(msg, messages.Event):
            endpoint = self._subscriptions[msg.subscription_id]
            try:
                await endpoint(types.Event(msg.args, msg.kwargs, msg.details))
            except Exception as e:
                print(e)
        elif isinstance(msg, messages.Error):
            match msg.message_type:
                case messages.Call.TYPE:
                    call_request = self._call_requests.pop(msg.request_id)
                    call_request.set_exception(exception_from_error(msg))
                case messages.Register.TYPE:
                    register_request = self._register_requests.pop(msg.request_id)
                    register_request.future.set_exception(exception_from_error(msg))
                case messages.Unregister.TYPE:
                    unregister_request = self._unregister_requests.pop(msg.request_id)
                    unregister_request.future.set_exception(exception_from_error(msg))
                case messages.Subscribe.TYPE:
                    subscribe_request = self._subscribe_requests.pop(msg.request_id)
                    subscribe_request.future.set_exception(exception_from_error(msg))
                case messages.Unsubscribe.TYPE:
                    unsubscribe_request = self._unsubscribe_requests.pop(msg.request_id)
                    unsubscribe_request.future.set_exception(exception_from_error(msg))
                case messages.Publish.TYPE:
                    publish_request = self._publish_requests.pop(msg.request_id)
                    publish_request.set_exception(exception_from_error(msg))
                case _:
                    raise exception.ProtocolError(msg.__str__())
        elif isinstance(msg, messages.Goodbye):
            self._goodbye_request.set_result(None)
        else:
            raise ValueError("received unknown message")

    async def register(
        self,
        procedure: str,
        invocation_handler: Callable[[types.Invocation], Awaitable[types.Result]],
        options: dict = None,
    ) -> Registration:
        if not inspect.iscoroutinefunction(invocation_handler):
            raise RuntimeError(
                f"function {invocation_handler.__name__} for procedure '{procedure}' must be a coroutine"
            )

        register = messages.Register(messages.RegisterFields(self._idgen.next(), procedure, options=options))
        data = self._session.send_message(register)

        f: Future[Registration] = Future()
        self._register_requests[register.request_id] = RegisterRequest(f, invocation_handler)
        await self._base_session.send(data)

        return await f

    async def call(
        self,
        procedure: str,
        args: list[Any] | None = None,
        kwargs: dict[str, Any] | None = None,
        options: dict[str, Any] | None = None,
    ) -> types.Result:
        call = messages.Call(messages.CallFields(self._idgen.next(), procedure, args, kwargs, options=options))
        data = self._session.send_message(call)

        f = Future()
        self._call_requests[call.request_id] = f

        await self._base_session.send(data)

        return await f

    async def subscribe(
        self, topic: str, event_handler: Callable[[types.Event], Awaitable[None]], options: dict | None = None
    ) -> Subscription:
        if not inspect.iscoroutinefunction(event_handler):
            raise RuntimeError(f"function {event_handler.__name__} for topic '{topic}' must be a coroutine")

        subscribe = messages.Subscribe(messages.SubscribeFields(self._idgen.next(), topic, options=options))
        data = self._session.send_message(subscribe)

        f: Future[Subscription] = Future()
        self._subscribe_requests[subscribe.request_id] = SubscribeRequest(f, event_handler)
        await self._base_session.send(data)

        return await f

    async def publish(
        self, topic: str, args: list[Any] | None = None, kwargs: dict | None = None, options: dict | None = None
    ) -> None:
        publish = messages.Publish(messages.PublishFields(self._idgen.next(), topic, args, kwargs, options))
        data = self._session.send_message(publish)

        if options is not None and options.get("acknowledge", False):
            f: Future = Future()
            self._publish_requests[publish.request_id] = f
            await self._base_session.send(data)
            return await f

        await self._base_session.send(data)

    async def leave(self) -> None:
        self._goodbye_request = Future()

        goodbye = messages.Goodbye(messages.GoodbyeFields({}, xconn_uris.CLOSE_REALM))
        data = self._session.send_message(goodbye)
        await self._base_session.send(data)

        try:
            await asyncio.wait_for(self._goodbye_request, timeout=10)
        except asyncio.TimeoutError:
            pass

        self.wait_task.cancel()
        try:
            await self.wait_task
        except asyncio.CancelledError:
            pass

        if await self._base_session.transport.is_connected():
            await self._base_session.close()

    async def ping(self, timeout: int = 10) -> float:
        return await self._base_session.transport.ping(timeout)

    def _on_disconnect(self, callback: Callable[[], Awaitable[None]]) -> None:
        if callback is not None:
            self._disconnect_callback.append(callback)



from __future__ import annotations

from concurrent.futures import Future
from threading import Thread
from typing import Callable, Any
from dataclasses import dataclass

from wampproto import messages, idgen, session, uris

from xconn import types, exception, uris as xconn_uris
from xconn.exception import ApplicationError
from xconn.helpers import exception_from_error


@dataclass
class RegisterRequest:
    future: Future[Registration]
    endpoint: Callable | Callable[[types.Invocation], types.Result]


class Registration:
    def __init__(self, registration_id: int, session: Session):
        self.registration_id = registration_id
        self._session = session

    def unregister(self) -> None:
        if not self._session._base_session.transport.is_connected():
            raise Exception("cannot unregister procedure: session not established")

        unregister = messages.Unregister(messages.UnregisterFields(self._session._idgen.next(), self.registration_id))
        data = self._session._session.send_message(unregister)

        f: Future = Future()
        self._session._unregister_requests[unregister.request_id] = types.UnregisterRequest(f, self.registration_id)
        self._session._base_session.send(data)

        f.result()


@dataclass
class SubscribeRequest:
    future: Future[Subscription]
    endpoint: Callable[[types.Event], None]


class Subscription:
    def __init__(self, subscription_id: int, session: Session):
        self.subscription_id = subscription_id
        self._session = session

    def unsubscribe(self) -> None:
        if not self._session._base_session.transport.is_connected():
            raise Exception("cannot unsubscribe topic: session not established")

        unsubscribe = messages.Unsubscribe(
            messages.UnsubscribeFields(self._session._idgen.next(), self.subscription_id)
        )
        data = self._session._session.send_message(unsubscribe)

        f: Future = Future()
        self._session._unsubscribe_requests[unsubscribe.request_id] = types.UnsubscribeRequest(f, self.subscription_id)
        self._session._base_session.send(data)

        f.result()


class Session:
    def __init__(self, base_session: types.BaseSession):
        # RPC data structures
        self._call_requests: dict[int, Future[types.Result]] = {}
        self._register_requests: dict[int, RegisterRequest] = {}
        self._registrations: dict[int, Callable[[types.Invocation], types.Result]] = {}
        self._unregister_requests: dict[int, types.UnregisterRequest] = {}

        # PubSub data structures
        self._publish_requests: dict[int, Future[None]] = {}
        self._subscribe_requests: dict[int, SubscribeRequest] = {}
        self._subscriptions: dict[int, Callable[[types.Event], None]] = {}
        self._unsubscribe_requests: dict[int, types.UnsubscribeRequest] = {}

        self._goodbye_request = Future()

        # ID generator
        self._idgen = idgen.SessionScopeIDGenerator()

        self._base_session = base_session

        # initialize the sans-io wamp session
        self._session = session.WAMPSession(base_session.serializer)

        self._disconnect_callback: list[Callable[[], None] | None] = []

        thread = Thread(target=self._wait, daemon=False)
        thread.start()

    def _wait(self):
        while self._base_session.transport.is_connected():
            try:
                data = self._base_session.receive()
            except Exception:
                break

            self._process_incoming_message(self._session.receive(data))

        for callback in self._disconnect_callback:
            callback()

    def _process_incoming_message(self, msg: messages.Message):
        if isinstance(msg, messages.Registered):
            request = self._register_requests.pop(msg.request_id)
            self._registrations[msg.registration_id] = request.endpoint
            request.future.set_result(Registration(msg.registration_id, self))
        elif isinstance(msg, messages.Unregistered):
            request = self._unregister_requests.pop(msg.request_id)
            del self._registrations[request.registration_id]
            request.future.set_result(None)
        elif isinstance(msg, messages.Result):
            request = self._call_requests.pop(msg.request_id)
            request.set_result(types.Result(msg.args, msg.kwargs, msg.details))
        elif isinstance(msg, messages.Invocation):
            try:
                endpoint = self._registrations[msg.registration_id]
                result = endpoint(types.Invocation(msg.args, msg.kwargs, msg.details))

                if result is None:
                    data = self._session.send_message(messages.Yield(messages.YieldFields(msg.request_id)))
                elif isinstance(result, types.Result):
                    data = self._session.send_message(
                        messages.Yield(messages.YieldFields(msg.request_id, result.args, result.kwargs, result.details))
                    )
                else:
                    message = "Endpoint returned invalid result type. Expected types.Result or None, got: " + str(
                        type(result)
                    )
                    msg_to_send = messages.Error(
                        messages.ErrorFields(msg.TYPE, msg.request_id, xconn_uris.ERROR_INTERNAL_ERROR, [message])
                    )
                    data = self._session.send_message(msg_to_send)

                self._base_session.send(data)
            except ApplicationError as e:
                msg_to_send = messages.Error(messages.ErrorFields(msg.TYPE, msg.request_id, e.message, e.args))
                data = self._session.send_message(msg_to_send)
                self._base_session.send(data)
            except Exception as e:
                msg_to_send = messages.Error(
                    messages.ErrorFields(msg.TYPE, msg.request_id, xconn_uris.ERROR_RUNTIME_ERROR, [e.__str__()])
                )
                data = self._session.send_message(msg_to_send)
                self._base_session.send(data)
        elif isinstance(msg, messages.Subscribed):
            request = self._subscribe_requests.pop(msg.request_id)
            self._subscriptions[msg.subscription_id] = request.endpoint
            request.future.set_result(Subscription(msg.subscription_id, self))
        elif isinstance(msg, messages.Unsubscribed):
            request = self._unsubscribe_requests.pop(msg.request_id)
            del self._subscriptions[request.subscription_id]
            request.future.set_result(None)
        elif isinstance(msg, messages.Published):
            request = self._publish_requests.pop(msg.request_id)
            request.set_result(None)
        elif isinstance(msg, messages.Event):
            try:
                endpoint = self._subscriptions[msg.subscription_id]
                endpoint(types.Event(msg.args, msg.kwargs, msg.details))
            except Exception as e:
                print(e)
        elif isinstance(msg, messages.Error):
            match msg.message_type:
                case messages.Call.TYPE:
                    call_request = self._call_requests.pop(msg.request_id)
                    call_request.set_exception(exception_from_error(msg))
                case messages.Register.TYPE:
                    register_request = self._register_requests.pop(msg.request_id)
                    register_request.future.set_exception(exception_from_error(msg))
                case messages.Unregister.TYPE:
                    unregister_request = self._unregister_requests.pop(msg.request_id)
                    unregister_request.future.set_exception(exception_from_error(msg))
                case messages.Subscribe.TYPE:
                    subscribe_request = self._subscribe_requests.pop(msg.request_id)
                    subscribe_request.future.set_exception(exception_from_error(msg))
                case messages.Unsubscribe.TYPE:
                    unsubscribe_request = self._unsubscribe_requests.pop(msg.request_id)
                    unsubscribe_request.future.set_exception(exception_from_error(msg))
                case messages.Publish.TYPE:
                    publish_request = self._publish_requests.pop(msg.request_id)
                    publish_request.set_exception(exception_from_error(msg))
                case _:
                    raise exception.ProtocolError(msg.__str__())
        elif isinstance(msg, messages.Goodbye):
            self._goodbye_request.set_result(None)
        else:
            raise ValueError("received unknown message")

    def call(
        self,
        procedure: str,
        args: list[Any] | None = None,
        kwargs: dict[str, Any] | None = None,
        options: dict[str, Any] | None = None,
    ) -> types.Result:
        if options is not None and options.get("x_payload_raw", False):
            options.pop("x_payload_raw", None)
            if len(args) > 1:
                raise TypeError("must provide at most one argument when 'x_payload_raw' is set")

            if len(kwargs) != 0:
                raise TypeError("must not provide kwargs when 'x_payload_raw' is set")

            if len(args) == 0:
                call = messages.Call(
                    messages.CallFields(self._idgen.next(), procedure, options=options, serializer=0, binary=True)
                )
            else:
                if not isinstance(args[0], bytearray):
                    raise TypeError("argument must be of type bytearray when 'x_payload_raw' is set")

                call = messages.Call(
                    messages.CallFields(
                        self._idgen.next(),
                        procedure,
                        options=options,
                        payload=args[0],
                        serializer=0,
                        binary=True,
                    )
                )
        else:
            call = messages.Call(messages.CallFields(self._idgen.next(), procedure, args, kwargs, options=options))

        data = self._session.send_message(call)

        f = Future()
        self._call_requests[call.request_id] = f
        self._base_session.send(data)

        return f.result()

    def register(
        self,
        procedure: str,
        invocation_handler: Callable | Callable[[types.Invocation], types.Result],
        options: dict = None,
    ) -> Registration:
        register = messages.Register(messages.RegisterFields(self._idgen.next(), procedure, options=options))
        data = self._session.send_message(register)

        f: Future[Registration] = Future()
        self._register_requests[register.request_id] = RegisterRequest(f, invocation_handler)

        self._base_session.send(data)

        return f.result()

    def subscribe(self, topic: str, event_handler: Callable[[types.Event], None], options: dict = None) -> Subscription:
        subscribe = messages.Subscribe(messages.SubscribeFields(self._idgen.next(), topic, options=options))
        data = self._session.send_message(subscribe)

        f: Future[Subscription] = Future()
        self._subscribe_requests[subscribe.request_id] = SubscribeRequest(f, event_handler)
        self._base_session.send(data)

        return f.result()

    def publish(self, topic: str, args: list[Any] = None, kwargs: dict = None, options: dict = None):
        publish = messages.Publish(messages.PublishFields(self._idgen.next(), topic, args, kwargs, options))
        data = self._session.send_message(publish)

        if options is not None and options.get("acknowledge", False):
            f: Future = Future()
            self._publish_requests[publish.request_id] = f
            self._base_session.send(data)
            return f.result()

        self._base_session.send(data)

    def leave(self):
        self._goodbye_request = Future()

        goodbye = messages.Goodbye(messages.GoodbyeFields({}, uris.CLOSE_REALM))
        data = self._session.send_message(goodbye)
        self._base_session.send(data)
        try:
            self._goodbye_request.result(timeout=10)
        finally:
            self._base_session.close()

    def ping(self, timeout: int = 10) -> float:
        return self._base_session.transport.ping(timeout)

    def _on_disconnect(self, callback: Callable[[], None]) -> None:
        if callback is not None:
            self._disconnect_callback.append(callback)



import asyncio
from typing import Callable, Type, Awaitable
import inspect
from enum import Enum

from pydantic import BaseModel

from xconn.types import RegisterOptions, SubscribeOptions
from xconn.client import Session
from xconn.async_session import AsyncSession


class ExecutionMode(Enum):
    ASYNC = "ASYNC"
    SYNC = "SYNC"
    UNIVERSAL = "UNIVERSAL"


class IComponent:
    def set_session(self, session: Session | AsyncSession):
        raise NotImplementedError()

    @property
    def session(self) -> Session | AsyncSession:
        raise NotImplementedError()

    @property
    def procedures(self) -> dict[str, Callable]:
        raise NotImplementedError()

    @property
    def topics(self) -> dict[str, Callable]:
        raise NotImplementedError()

    def register(self, procedure: str):
        raise NotImplementedError()

    def subscribe(self, topic: str):
        raise NotImplementedError()


def register(
    procedure: str,
    response_model: Type[BaseModel] | None = None,
    allowed_roles: list[str] = None,
    options: dict | RegisterOptions | None = None,
):
    def _register(func):
        func.__xconn_procedure__ = procedure
        func.__xconn_response_model__ = response_model
        func.__xconn_allowed_roles__ = allowed_roles if isinstance(allowed_roles, list) else []
        func.__xconn_register_options__ = options
        return func

    return _register


def subscribe(topic: str, options: dict | SubscribeOptions | None = None):
    def _subscribe(func):
        func.__xconn_topic__ = topic
        func.__xconn_subscribe_options__ = options
        return func

    return _subscribe


class Component(IComponent):
    def __init__(self):
        super().__init__()
        self._procedures = {}
        self._topics = {}

        self._session: Session | AsyncSession = None

        for name, method in inspect.getmembers(self.__class__, inspect.isfunction):
            if getattr(method, "__xconn_procedure__", False):
                self._procedures[name] = method.__get__(self)
            elif getattr(method, "__xconn_topic__", False):
                self._topics[name] = method.__get__(self)

    def set_session(self, session: Session | AsyncSession):
        self._session = session

    @property
    def session(self) -> Session | AsyncSession:
        return self._session

    @property
    def procedures(self) -> dict[str, Callable]:
        return self._procedures

    @property
    def topics(self) -> dict[str, Callable]:
        return self._topics

    def register(
        self,
        procedure: str,
        response_model: Type[BaseModel] | None = None,
        allowed_roles: list[str] = None,
        options: dict | RegisterOptions | None = None,
    ):
        def _register(func):
            if procedure in self._procedures:
                raise ValueError(f"procedure {procedure} already registered")

            if response_model is not None:
                if not issubclass(response_model, BaseModel):
                    raise ValueError(f"response_model {response_model} is not a subclass of BaseModel")

            func.__xconn_response_model__ = response_model
            func.__xconn_allowed_roles__ = allowed_roles if isinstance(allowed_roles, list) else []
            func.__xconn_register_options__ = options
            self._procedures[procedure] = func

        return _register

    def subscribe(self, topic: str, options: dict | SubscribeOptions | None = None):
        def _subscribe(func):
            if topic in self._topics:
                raise ValueError(f"topic {topic} already registered")

            func.__xconn_subscribe_options__ = options
            self._topics[topic] = func

        return _subscribe


class App(Component):
    def __init__(self):
        super().__init__()
        self._procedures = {}
        self._topics = {}

        self._session: Session | AsyncSession | None = None
        self._components: list[Component] = []

        self._startup_handler: Callable | Awaitable[None] | None = None
        self._schema_procedure: str | None = None

        self._execution_mode: ExecutionMode = ExecutionMode.ASYNC

    def set_session(self, session: Session | AsyncSession):
        self._session = session
        for component in self._components:
            component.set_session(session)

        if self._startup_handler is not None:
            if inspect.iscoroutinefunction(self._startup_handler):
                asyncio.create_task(self._startup_handler())
            else:
                self._startup_handler()

    @property
    def components(self) -> list[Component]:
        return self._components

    @property
    def schema_procedure(self) -> str:
        return self._schema_procedure

    @property
    def execution_mode(self) -> ExecutionMode:
        return self._execution_mode

    def include_component(self, component: Component, prefix: str = "") -> None:
        if prefix is None or len(prefix) == 0:
            self._procedures.update(component.procedures)
            self._topics.update(component.topics)
        else:
            for procedure, func in component.procedures.items():
                self._procedures.update({prefix + procedure: func})

            for topic, func in component.topics.items():
                self._topics.update({prefix + topic: func})

        self._components.append(component)

    def add_event_handler(self, event_type: str, handler: Callable | Awaitable):
        if event_type != "startup":
            raise ValueError(f"event_type {event_type} is not supported")

        self._startup_handler = handler

    def set_schema_procedure(self, uri: str):
        self._schema_procedure = uri

    def set_execution_mode(self, mode: ExecutionMode):
        self._execution_mode = mode



import contextlib
import random
import inspect
import threading
import time
from typing import Any, Generator
from urllib.parse import urlparse

from xconn import App
from xconn._client.helpers import (
    _validate_procedure_function,
    _validate_topic_function,
    _handle_result,
    _sanitize_incoming_data,
    collect_docs,
    select_authenticator,
    start_server_sync,
    wait_for_server,
    handle_model_validation,
    ensure_caller_allowed,
    INITIAL_WAIT,
    MAX_WAIT,
    ProcedureMetadata,
    assemble_call_details,
    assemble_event_details,
    validate_invocation_parameters,
    validate_event_parameters,
    import_app,
)
from xconn._client.types import ClientConfig
from xconn.client import Client
from xconn.session import Session
from xconn.types import Event, Invocation, Result, RegisterOptions, InvokeOptions


def _setup(app: App, session: Session):
    app.set_session(session)

    for uri, func in app.procedures.items():
        register_sync(session, uri, func)

    for uri, func in app.topics.items():
        subscribe_sync(session, uri, func)


def _connect_sync(app: App, config: ClientConfig, start_router: bool = False, directory: str = "."):
    ws_url = urlparse(config.url)

    if start_router:
        threading.Thread(target=start_server_sync, args=(config,), daemon=True).start()
        wait_for_server(ws_url.hostname, ws_url.port, 5)

    auth = select_authenticator(config)
    client = Client(authenticator=auth, ws_config=config.websocket_config)

    def wait_and_connect(previous_wait: float = INITIAL_WAIT):
        next_wait = min(random.uniform(INITIAL_WAIT, previous_wait * 3), MAX_WAIT)
        print(f"reconnecting in {next_wait:.1f} seconds...")
        time.sleep(next_wait)

        try:
            new_session = client.connect(config.url, config.realm, on_connect, on_disconnect)
        except Exception as e:
            print(e)

            wait_and_connect(next_wait)
            return

        _setup(app, new_session)

    def on_connect():
        print("connected", config.realm)

    def on_disconnect():
        print("disconnected", config.realm)
        wait_and_connect()

    session = client.connect(config.url, config.realm, on_connect, on_disconnect)
    _setup(app, session)

    if app.schema_procedure is not None and app.schema_procedure != "":
        docs = []

        for uri, func in app.procedures.items():
            docs.append(collect_docs(uri, func, "procedure"))

        for uri, func in app.topics.items():
            docs.append(collect_docs(uri, func, "topic"))

        def get_schema(_: Invocation) -> Result:
            return Result(args=docs)

        options = RegisterOptions(invoke=InvokeOptions.ROUNDROBIN)
        session.register(app.schema_procedure, get_schema, options=options)
        print(f"serving schema at procedure {app.schema_procedure}")


def connect_sync(app: str, config: ClientConfig, start_router: bool = False, directory: str = "."):
    imported_app = import_app(app, directory)
    _connect_sync(imported_app, config, start_router, directory)


@contextlib.contextmanager
def resolve_dependencies(meta: ProcedureMetadata) -> Generator[dict[Any, Any], Any, None]:
    result = {}

    for key, value in meta.dependencies.items():
        result[key] = value()

    with contextlib.ExitStack() as stack:
        for name, dependency in meta.ctx_dependencies.items():
            result[name] = stack.enter_context(dependency())

        yield result


def register_sync(session: Session, uri: str, func: callable):
    if inspect.iscoroutinefunction(func):
        raise RuntimeError(f"function {func.__name__} for procedure '{uri}' must not be a coroutine")

    meta = _validate_procedure_function(func, uri)

    def _handle_invocation(invocation: Invocation) -> Result:
        ensure_caller_allowed(invocation.details, meta.allowed_roles)
        validate_invocation_parameters(invocation, meta)
        details = assemble_call_details(uri, meta, invocation)

        if meta.dynamic_model:
            kwargs = _sanitize_incoming_data(invocation.args, invocation.kwargs, meta.request_args)
            handle_model_validation(meta.request_model, **kwargs)

            with resolve_dependencies(meta) as deps:
                result = func(**kwargs, **deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)
        elif meta.request_model is not None:
            kwargs = _sanitize_incoming_data(invocation.args, invocation.kwargs, meta.request_args)
            model = handle_model_validation(meta.request_model, **kwargs)

            with resolve_dependencies(meta) as deps:
                input_data = {meta.positional_field_name: model}
                result = func(**input_data, **deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)
        elif meta.no_args:
            with resolve_dependencies(meta) as deps:
                result = func(**deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)
        else:
            with resolve_dependencies(meta) as deps:
                input_data = {meta.positional_field_name: invocation}
                result = func(**input_data, **deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)

    session.register(uri, _handle_invocation, options=meta.register_options)
    print(f"Registered procedure {uri}")


def subscribe_sync(session: Session, topic: str, func: callable):
    if inspect.iscoroutinefunction(func):
        raise RuntimeError(f"function {func.__name__} for topic '{topic}' must not be a coroutine")

    meta = _validate_topic_function(func, topic)

    def _handle_event(event: Event):
        details = assemble_event_details(topic, meta, event)
        validate_event_parameters(event, meta)

        if meta.dynamic_model:
            kwargs = _sanitize_incoming_data(event.args, event.kwargs, meta.request_args)
            handle_model_validation(meta.request_model, **kwargs)

            with resolve_dependencies(meta) as deps:
                func(**kwargs, **deps, **details)

        elif meta.request_model is not None:
            kwargs = _sanitize_incoming_data(event.args, event.kwargs, meta.request_args)
            model = handle_model_validation(meta.request_model, **kwargs)

            with resolve_dependencies(meta) as deps:
                input_data = {meta.positional_field_name: model}
                func(**input_data, **deps, **details)

        elif meta.no_args:
            with resolve_dependencies(meta) as deps:
                func(**deps, **details)
        else:
            with resolve_dependencies(meta) as deps:
                input_data = {meta.positional_field_name: event}
                func(**input_data, **deps, **details)

    session.subscribe(topic, _handle_event, options=meta.subscribe_options)
    print(f"Subscribed topic {topic}")



from wampproto import serializers
from wampproto.messages import Error
from wampproto.transports.rawsocket import SERIALIZER_TYPE_JSON, SERIALIZER_TYPE_MSGPACK, SERIALIZER_TYPE_CBOR

try:
    from wamp_msgs_capnp.serializer import CapnProtoSerializer

    _CAPNP_AVAILABLE = True
except ImportError:
    _CAPNP_AVAILABLE = False

from xconn.exception import ApplicationError

JSON_SUBPROTOCOL = "wamp.2.json"
CBOR_SUBPROTOCOL = "wamp.2.cbor"
MSGPACK_SUBPROTOCOL = "wamp.2.msgpack"
CAPNPROTO_SUBPROTOCOL = "wamp.2.capnproto.split_payload"

SERIALIZER_TYPE_CAPNPROTO = 14


def get_ws_subprotocol(serializer: serializers.Serializer):
    if isinstance(serializer, serializers.JSONSerializer):
        return JSON_SUBPROTOCOL
    elif isinstance(serializer, serializers.CBORSerializer):
        return CBOR_SUBPROTOCOL
    elif isinstance(serializer, serializers.MsgPackSerializer):
        return MSGPACK_SUBPROTOCOL
    elif _CAPNP_AVAILABLE and isinstance(serializer, CapnProtoSerializer):
        return CAPNPROTO_SUBPROTOCOL
    elif isinstance(serializer, type) and serializer.__name__ == "CapnProtoSerializer":
        raise ImportError(
            "Cap'n Proto serializer support is not installed.\nInstall it with:\n  uv pip install xconn[capnproto]"
        )
    else:
        raise ValueError("invalid serializer")


def get_rs_protocol(serializer: serializers.Serializer):
    if isinstance(serializer, serializers.JSONSerializer):
        return SERIALIZER_TYPE_JSON
    elif isinstance(serializer, serializers.CBORSerializer):
        return SERIALIZER_TYPE_CBOR
    elif isinstance(serializer, serializers.MsgPackSerializer):
        return SERIALIZER_TYPE_MSGPACK
    elif _CAPNP_AVAILABLE and isinstance(serializer, CapnProtoSerializer):
        return SERIALIZER_TYPE_CAPNPROTO
    elif isinstance(serializer, type) and serializer.__name__ == "CapnProtoSerializer":
        raise ImportError(
            "Cap'n Proto serializer support is not installed.\nInstall it with:\n  uv pip install xconn[capnproto]"
        )

    else:
        raise ValueError("invalid serializer")


def get_serializer(ws_subprotocol: str) -> serializers.Serializer:
    if ws_subprotocol == JSON_SUBPROTOCOL:
        return serializers.JSONSerializer()
    elif ws_subprotocol == CBOR_SUBPROTOCOL:
        return serializers.CBORSerializer()
    elif ws_subprotocol == MSGPACK_SUBPROTOCOL:
        return serializers.MsgPackSerializer()
    elif ws_subprotocol == CAPNPROTO_SUBPROTOCOL:
        if not _CAPNP_AVAILABLE:
            raise ImportError(
                "Cap'n Proto serializer support is not installed.\nInstall it with:\n  uv pip install xconn[capnproto]"
            )
        return CapnProtoSerializer()
    else:
        raise ValueError(f"invalid websocket subprotocol {ws_subprotocol}")


def exception_from_error(error: Error):
    exc = ApplicationError(error.uri)
    if error.args:
        exc.args = error.args
    if error.kwargs:
        exc.kwargs = error.kwargs

    return exc



class ApplicationError(RuntimeError):
    def __init__(self, message: str, *args, **kwargs):
        super().__init__(*args)
        self.message = message
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        err = self.message

        if self.args:
            args = ", ".join(str(arg) for arg in self.args)
            err += f"{args}"

        if self.kwargs:
            kwargs = ", ".join(f"{key}={value}" for key, value in self.kwargs.items())
            err += f": {kwargs}"

        return err


class ProtocolError(Exception):
    def __init__(self, message: str):
        super().__init__(message)



import contextlib
import random
import asyncio
import inspect
from typing import AsyncGenerator

from xconn import App, AsyncClient
from xconn._client.helpers import (
    _validate_procedure_function,
    _validate_topic_function,
    _handle_result,
    _sanitize_incoming_data,
    collect_docs,
    select_authenticator,
    start_server_async,
    handle_model_validation,
    ensure_caller_allowed,
    INITIAL_WAIT,
    MAX_WAIT,
    ProcedureMetadata,
    assemble_call_details,
    assemble_event_details,
    validate_invocation_parameters,
    validate_event_parameters,
    import_app,
)
from xconn._client.types import ClientConfig
from xconn.async_session import AsyncSession
from xconn.types import Event, Invocation, Result, RegisterOptions, InvokeOptions
from xconn.utils import run


async def _setup(app: App, session: AsyncSession):
    app.set_session(session)

    for uri, func in app.procedures.items():
        await register_async(session, uri, func)

    for uri, func in app.topics.items():
        await subscribe_async(session, uri, func)


async def _connect_async(app: App, config: ClientConfig, start_router: bool = False):
    if start_router:
        await start_server_async(config)

    auth = select_authenticator(config)
    client = AsyncClient(authenticator=auth, ws_config=config.websocket_config)

    async def wait_and_connect(previous_wait: float = INITIAL_WAIT):
        next_wait = min(random.uniform(INITIAL_WAIT, previous_wait * 3), MAX_WAIT)
        print(f"reconnecting in {next_wait:.1f} seconds...")
        await asyncio.sleep(next_wait)

        try:
            new_session = await client.connect(config.url, config.realm, on_connect, on_disconnect)
        except Exception as e:
            print(e)

            await wait_and_connect(next_wait)
            return

        await _setup(app, new_session)

    async def on_connect():
        print("connected", config.realm)

    async def on_disconnect():
        print("disconnected", config.realm)
        await wait_and_connect()

    session = await client.connect(config.url, config.realm, on_connect, on_disconnect)
    await _setup(app, session)

    if app.schema_procedure is not None and app.schema_procedure != "":
        docs = []

        for uri, func in app.procedures.items():
            docs.append(collect_docs(uri, func, "procedure"))

        for uri, func in app.topics.items():
            docs.append(collect_docs(uri, func, "topic"))

        async def get_schema(_: Invocation) -> Result:
            return Result(args=docs)

        options = RegisterOptions(invoke=InvokeOptions.ROUNDROBIN)
        await session.register(app.schema_procedure, get_schema, options=options)
        print(f"serving schema at procedure {app.schema_procedure}")


def connect_async(app: str, config: ClientConfig, start_router: bool = False, directory: str = "."):
    app = import_app(app, directory)
    # FIXME: also support running from an existing event loop instead of always starting a new one.
    run(_connect_async(app, config, start_router=start_router))


@contextlib.asynccontextmanager
async def resolve_dependencies(meta: ProcedureMetadata) -> AsyncGenerator:
    result = {}

    for key, value in meta.async_dependencies.items():
        result[key] = await value()

    async with contextlib.AsyncExitStack() as stack:
        for name, dependency in meta.async_ctx_dependencies.items():
            result[name] = await stack.enter_async_context(dependency())

        yield result


async def register_async(session: AsyncSession, uri: str, func: callable):
    if not inspect.iscoroutinefunction(func):
        raise RuntimeError(f"function {func.__name__} for procedure '{uri}' must be a coroutine")

    meta = _validate_procedure_function(func, uri)

    async def _handle_invocation(invocation: Invocation) -> Result:
        ensure_caller_allowed(invocation.details, meta.allowed_roles)
        validate_invocation_parameters(invocation, meta)
        details = assemble_call_details(uri, meta, invocation)

        if meta.dynamic_model:
            kwargs = _sanitize_incoming_data(invocation.args, invocation.kwargs, meta.request_args)
            handle_model_validation(meta.request_model, **kwargs)

            async with resolve_dependencies(meta) as deps:
                result = await func(**kwargs, **deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)
        elif meta.request_model is not None:
            kwargs = _sanitize_incoming_data(invocation.args, invocation.kwargs, meta.request_args)
            model = handle_model_validation(meta.request_model, **kwargs, **details)
            input_data = {meta.positional_field_name: model}

            async with resolve_dependencies(meta) as deps:
                result = await func(**input_data, **deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)
        elif meta.no_args:
            async with resolve_dependencies(meta) as deps:
                result = await func(**deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)
        else:
            async with resolve_dependencies(meta) as deps:
                input_data = {meta.positional_field_name: invocation}
                result = await func(**input_data, **deps, **details)

            return _handle_result(result, meta.response_model, meta.response_args)

    await session.register(uri, _handle_invocation, options=meta.register_options)
    print(f"Registered procedure {uri}")


async def subscribe_async(session: AsyncSession, topic: str, func: callable):
    if not inspect.iscoroutinefunction(func):
        raise RuntimeError(f"function {func.__name__} for topic '{topic}' must be a coroutine")

    meta = _validate_topic_function(func, topic)

    async def _handle_event(event: Event) -> None:
        details = assemble_event_details(topic, meta, event)
        validate_event_parameters(event, meta)

        if meta.dynamic_model:
            kwargs = _sanitize_incoming_data(event.args, event.kwargs, meta.request_args)
            handle_model_validation(meta.request_model, **kwargs)

            async with resolve_dependencies(meta) as deps:
                await func(**kwargs, **deps, **details)

        elif meta.request_model is not None:
            kwargs = _sanitize_incoming_data(event.args, event.kwargs, meta.request_args)
            model = handle_model_validation(meta.request_model, **kwargs)

            async with resolve_dependencies(meta) as deps:
                input_data = {meta.positional_field_name: model}
                await func(**input_data, **deps, **details)

        elif meta.no_args:
            async with resolve_dependencies(meta) as deps:
                await func(**deps, **details)
        else:
            async with resolve_dependencies(meta) as deps:
                input_data = {meta.positional_field_name: event}
                await func(**input_data, **deps, **details)

    await session.subscribe(topic, _handle_event, options=meta.subscribe_options)
    print(f"Subscribed topic {topic}")



from wampproto import messages

from xconn import realm, types


class Router:
    def __init__(self):
        super().__init__()
        self.realms: dict[str, realm.Realm] = {}

    def add_realm(self, name: str):
        self.realms[name] = realm.Realm()

    def remove_realm(self, name: str):
        del self.realms[name]

    def has_realm(self, name: str):
        return name in self.realms

    def attach_client(self, base_session: types.IAsyncBaseSession):
        if base_session.realm not in self.realms:
            raise ValueError(f"cannot attach client to non-existent realm {base_session.realm}")

        self.realms[base_session.realm].attach_client(base_session)

    def detach_client(self, base_session: types.IAsyncBaseSession):
        if base_session.realm not in self.realms:
            raise ValueError(f"cannot detach client from non-existent realm {base_session.realm}")

        self.realms[base_session.realm].detach_client(base_session)

    async def receive_message(self, base_session: types.IAsyncBaseSession, msg: messages.Message):
        if base_session.realm not in self.realms:
            raise ValueError(f"cannot process message for non-existent realm {base_session.realm}")

        await self.realms[base_session.realm].receive_message(base_session.id, msg)



from typing import Callable
from urllib.parse import urlparse

from wampproto import auth, serializers

from xconn import types
from xconn.session import Session
from xconn.joiner import WebsocketsJoiner, RawSocketJoiner


class Client:
    def __init__(
        self,
        authenticator: auth.IClientAuthenticator = auth.AnonymousAuthenticator(""),
        serializer: serializers.Serializer = serializers.JSONSerializer(),
        ws_config: types.WebsocketConfig = types.WebsocketConfig(),
    ):
        self._authenticator = authenticator
        self._serializer = serializer
        self._ws_config = ws_config

    def connect(
        self,
        uri: str,
        realm: str,
        connect_callback: Callable[[], None] | None = None,
        disconnect_callback: Callable[[], None] | None = None,
    ) -> Session:
        return connect(
            uri, realm, self._authenticator, self._serializer, self._ws_config, connect_callback, disconnect_callback
        )


def connect(
    uri: str,
    realm: str,
    authenticator: auth.IClientAuthenticator = None,
    serializer: serializers.Serializer = serializers.CBORSerializer(),
    ws_config: types.WebsocketConfig = types.WebsocketConfig(),
    connect_callback: Callable[[], None] | None = None,
    disconnect_callback: Callable[[], None] | None = None,
) -> Session:
    parsed = urlparse(uri)
    if parsed.scheme == "ws" or parsed.scheme == "wss" or parsed.scheme == "unix+ws":
        j = WebsocketsJoiner(authenticator, serializer, ws_config)
    elif (
        parsed.scheme == "rs"
        or parsed.scheme == "rss"
        or parsed.scheme == "tcp"
        or parsed.scheme == "tcps"
        or parsed.scheme == "unix"
        or parsed.scheme == "unix+rs"
    ):
        j = RawSocketJoiner(authenticator, serializer)
    else:
        raise RuntimeError(f"Unsupported scheme {parsed.scheme}")

    details = j.join(uri, realm)
    session = Session(details)

    session._on_disconnect(disconnect_callback)

    if connect_callback is not None:
        connect_callback()

    return session


def connect_anonymous(uri: str, realm: str) -> Session:
    return connect(uri, realm)


def connect_ticket(uri: str, realm: str, authid: str, ticket: str) -> Session:
    ticket_authenticator = auth.TicketAuthenticator(authid, ticket)

    return connect(uri, realm, ticket_authenticator)


def connect_wampcra(uri: str, realm: str, authid: str, secret: str) -> Session:
    wampcra_authenticator = auth.WAMPCRAAuthenticator(authid, secret)

    return connect(uri, realm, wampcra_authenticator)


def connect_cryptosign(uri: str, realm: str, authid: str, private_key: str) -> Session:
    cryptosign_authenticator = auth.CryptoSignAuthenticator(authid, private_key)

    return connect(uri, realm, cryptosign_authenticator)



import enum

from pydantic import BaseModel, Field

from xconn.types import WebsocketConfig


class ConfigSource(str, enum.Enum):
    env = "env"
    cli = "cli"
    file = "file"


class ClientConfig(BaseModel):
    url: str
    realm: str
    websocket_config: WebsocketConfig = WebsocketConfig()

    authid: str | None = None
    authmethod: str | None = None
    secret: str | None = None
    ticket: str | None = None
    private_key: str | None = None


class CommandArgs(BaseModel):
    app: str = Field(alias="APP")
    url: str | None = None
    realm: str | None = None
    directory: str | None = None
    config_source: str = ConfigSource.cli
    config_file: str | None = None
    start_router: bool
    reload: bool

    open_timeout: int | None = None
    ping_interval: int | None = None
    ping_timeout: int | None = None

    authid: str | None = None
    ticket: str | None = None
    secret: str | None = None
    private_key: str | None = None



from __future__ import annotations

import asyncio
import contextlib
import inspect
from asyncio import Future
from collections import deque
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Awaitable

from aiohttp import web
from wampproto import messages, joiner, serializers


@dataclass
class UnregisterRequest:
    future: Future
    registration_id: int


@dataclass
class UnsubscribeRequest:
    future: Future
    subscription_id: int


@dataclass
class Result:
    args: list | None = None
    kwargs: dict | None = None
    details: dict | None = None


@dataclass
class Invocation:
    args: list | None
    kwargs: dict | None
    details: dict | None


@dataclass
class Event:
    args: list | None
    kwargs: dict | None
    details: dict | None


@dataclass
class WebsocketConfig:
    # max wait time for connection to be established
    open_timeout: float | None = 10

    # send ping automatically after every x seconds
    ping_interval: float | None = 20

    # wait for x seconds for a pong from server before closing the connection
    ping_timeout: float | None = 20

    # max wait time for closing the connection
    close_timeout: float | None = 10


class ITransport:
    def read(self) -> str | bytes:
        raise NotImplementedError()

    def write(self, data: str | bytes):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def is_connected(self) -> bool:
        raise NotImplementedError()

    def ping(self, timeout: int = 10) -> float:
        raise NotImplementedError()


class IAsyncTransport:
    async def read(self) -> str | bytes:
        raise NotImplementedError()

    async def write(self, data: str | bytes):
        raise NotImplementedError()

    async def close(self):
        raise NotImplementedError()

    async def is_connected(self) -> bool:
        raise NotImplementedError()

    async def ping(self, timeout: int = 10) -> float:
        raise NotImplementedError()


class IBasePeer:
    def read(self) -> str | bytes:
        raise NotImplementedError()

    def write(self, data: str | bytes):
        raise NotImplementedError()


class BasePeer(IBasePeer):
    def read(self) -> str | bytes:
        return super().read()

    def write(self, data: str | bytes):
        super().write(data)


class IPeer:
    def read(self) -> str | bytes:
        raise NotImplementedError()

    def write(self, data: str | bytes):
        raise NotImplementedError()

    def read_message(self) -> messages.Message:
        raise NotImplementedError()

    def write_message(self, msg: messages.Message):
        raise NotImplementedError()


class Peer(IPeer):
    def __init__(self, base_peer: BasePeer):
        super().__init__()
        self._base_peer = base_peer


class IBaseSession:
    @property
    def transport(self) -> ITransport:
        raise NotImplementedError()

    @property
    def id(self) -> int:
        raise NotImplementedError()

    @property
    def realm(self) -> str:
        raise NotImplementedError()

    @property
    def authid(self) -> str:
        raise NotImplementedError()

    @property
    def authrole(self) -> str:
        raise NotImplementedError()

    def send(self, data: bytes):
        raise NotImplementedError()

    def receive(self) -> bytes:
        raise NotImplementedError()

    def send_message(self, msg: messages.Message):
        raise NotImplementedError()

    def receive_message(self) -> messages.Message:
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


class BaseSession(IBaseSession):
    def __init__(
        self, transport: ITransport, session_details: joiner.SessionDetails, serializer: serializers.Serializer
    ):
        super().__init__()
        self._transport = transport
        self.session_details = session_details
        self.serializer = serializer

    @property
    def transport(self) -> ITransport:
        return self._transport

    @property
    def id(self) -> int:
        return self.session_details.session_id

    @property
    def realm(self) -> str:
        return self.session_details.realm

    @property
    def authid(self) -> str:
        return self.session_details.authid

    @property
    def authrole(self) -> str:
        return self.session_details.authrole

    def send(self, data: bytes):
        self._transport.write(data)

    def receive(self) -> bytes:
        return self._transport.read()

    def send_message(self, msg: messages.Message):
        self.send(self.serializer.serialize(msg))

    def receive_message(self) -> messages.Message:
        return self.serializer.deserialize(self.receive())

    def close(self):
        self._transport.close()


class IAsyncBaseSession:
    @property
    def transport(self) -> IAsyncTransport:
        raise NotImplementedError()

    @property
    def id(self) -> int:
        raise NotImplementedError()

    @property
    def realm(self) -> str:
        raise NotImplementedError()

    @property
    def authid(self) -> str:
        raise NotImplementedError()

    @property
    def authrole(self) -> str:
        raise NotImplementedError()

    @property
    def serializer(self) -> serializers.Serializer:
        raise NotImplementedError()

    async def send(self, data: bytes | str):
        raise NotImplementedError()

    async def receive(self) -> bytes | str:
        raise NotImplementedError()

    async def send_message(self, msg: messages.Message):
        raise NotImplementedError()

    async def receive_message(self) -> messages.Message:
        raise NotImplementedError()

    async def close(self):
        raise NotImplementedError()


class AsyncBaseSession(IAsyncBaseSession):
    def __init__(
        self, transport: IAsyncTransport, session_details: joiner.SessionDetails, serializer: serializers.Serializer
    ):
        super().__init__()
        self._transport = transport
        self.session_details = session_details
        self._serializer = serializer

    @property
    def transport(self) -> IAsyncTransport:
        return self._transport

    @property
    def id(self) -> int:
        return self.session_details.session_id

    @property
    def realm(self) -> str:
        return self.session_details.realm

    @property
    def authid(self) -> str:
        return self.session_details.authid

    @property
    def authrole(self) -> str:
        return self.session_details.authrole

    @property
    def serializer(self) -> serializers.Serializer:
        return self._serializer

    async def send(self, data: bytes):
        return await self._transport.write(data)

    async def receive(self) -> bytes:
        return await self._transport.read()

    async def send_message(self, msg: messages.Message):
        await self.send(self.serializer.serialize(msg))

    async def receive_message(self) -> messages.Message:
        return self.serializer.deserialize(await self.receive())

    async def close(self):
        await self._transport.close()


class AIOHttpBaseSession(IAsyncBaseSession):
    def __init__(
        self, ws: web.WebSocketResponse, session_details: joiner.SessionDetails, serializer: serializers.Serializer
    ):
        super().__init__()
        self.ws = ws
        self.session_details = session_details
        self._serializer = serializer
        if serializer is None or isinstance(serializer, serializers.JSONSerializer):
            self._send_func = ws.send_str
            self._receive_func = ws.receive_str
        else:
            self._send_func = ws.send_bytes
            self._receive_func = ws.receive_bytes

    @property
    def id(self) -> int:
        return self.session_details.session_id

    @property
    def realm(self) -> str:
        return self.session_details.realm

    @property
    def authid(self) -> str:
        return self.session_details.authid

    @property
    def authrole(self) -> str:
        return self.session_details.authrole

    @property
    def serializer(self) -> serializers.Serializer:
        return self._serializer

    async def send(self, data: bytes):
        await self._send_func(data)

    async def receive(self) -> bytes:
        return await self._receive_func()

    async def send_message(self, msg: messages.Message):
        await self.send(self.serializer.serialize(msg))

    async def receive_message(self) -> messages.Message:
        return self.serializer.deserialize(await self.receive())

    async def close(self):
        await self.ws.close()


class ClientSideLocalBaseSession(IAsyncBaseSession):
    def __init__(self, sid: int, realm: str, authid: str, authrole: str, serializer: serializers.Serializer, router):
        super().__init__()
        self._sid: int = sid
        self._realm: str = realm
        self._authid: str = authid
        self._authrole: str = authrole
        self._serializer: serializers.Serializer = serializer
        self._router = router

        self._incoming_messages = deque()
        self._cond = asyncio.Condition()

    @property
    def id(self) -> int:
        return self._sid

    @property
    def realm(self) -> str:
        return self._realm

    @property
    def authid(self) -> str:
        return self._authid

    @property
    def authrole(self) -> str:
        return self._authrole

    @property
    def serializer(self) -> serializers.Serializer:
        return self._serializer

    async def send(self, data: bytes | str):
        await self.send_message(self.serializer.deserialize(data))

    async def receive(self) -> bytes | str:
        async with self._cond:
            while not self._incoming_messages:
                await self._cond.wait()

            return self._incoming_messages.popleft()

    async def send_message(self, msg: messages.Message):
        await self._router.receive_message(self, msg)

    async def receive_message(self) -> messages.Message:
        return self.serializer.deserialize(await self.receive())

    async def close(self):
        pass

    async def feed(self, data: bytes | str):
        async with self._cond:
            self._incoming_messages.append(data)
            self._cond.notify()


class ServerSideLocalBaseSession(IAsyncBaseSession):
    def __init__(self, sid: int, realm: str, authid: str, authrole: str, serializer: serializers.Serializer):
        super().__init__()
        self._sid: int = sid
        self._realm: str = realm
        self._authid: str = authid
        self._authrole: str = authrole
        self._serializer: serializers.Serializer = serializer
        self._other: ClientSideLocalBaseSession = None

    def set_other(self, other: ClientSideLocalBaseSession):
        self._other = other

    @property
    def id(self) -> int:
        return self._sid

    @property
    def realm(self) -> str:
        return self._realm

    @property
    def authid(self) -> str:
        return self._authid

    @property
    def authrole(self) -> str:
        return self._authrole

    @property
    def serializer(self) -> serializers.Serializer:
        return self._serializer

    async def send(self, data: bytes | str):
        await self._other.feed(data)

    async def send_message(self, msg: messages.Message):
        await self.send(self.serializer.serialize(msg))

    async def close(self):
        pass


class _IncomingDetails(dict):
    def __init__(self, details: dict | None = None):
        super().__init__()
        if details:
            for k, v in details.items():
                self[k] = v


class CallDetails(_IncomingDetails):
    def __init__(self, details: dict | None = None):
        super().__init__(details)

    @property
    def session_id(self) -> int | None:
        return self.get("caller")

    @property
    def authid(self) -> str | None:
        return self.get("caller_authid")

    @property
    def authrole(self) -> str | None:
        return self.get("caller_authrole")


class EventDetails(_IncomingDetails):
    def __init__(self, details: dict | None = None):
        super().__init__(details)

    @property
    def session_id(self) -> int | None:
        return self.get("publisher")

    @property
    def authid(self) -> str | None:
        return self.get("publisher_authid")

    @property
    def authrole(self) -> str | None:
        return self.get("publisher_authrole")


class InvokeOptions(Enum):
    SINGLE = "single"
    ROUNDROBIN = "roundrobin"
    RANDOM = "random"
    FIRST = "first"
    LAST = "last"


class MatchOptions(Enum):
    EXACT = "exact"
    PREFIX = "prefix"
    WILDCARD = "wildcard"


class RegisterOptions(dict):
    def __init__(self, invoke: InvokeOptions = None, match: MatchOptions = None, concurrency: int = None, **kwargs):
        super().__init__()
        if invoke is not None:
            if not isinstance(invoke, InvokeOptions):
                raise ValueError("expected InvokeOptions for 'invoke' WAMP option")

            self["invoke"] = invoke.value
        if match is not None:
            if not isinstance(match, MatchOptions):
                raise ValueError("expected MatchOptions for 'match' WAMP option")

            self["match"] = match.value

        if concurrency is not None:
            if not isinstance(concurrency, int):
                raise ValueError("expected int for 'concurrency' WAMP option")

            self["concurrency"] = concurrency

        for k, v in kwargs.items():
            self[k] = v


class SubscribeOptions(dict):
    def __init__(self, match: MatchOptions = None, **kwargs):
        super().__init__()
        if match is not None:
            if not isinstance(match, MatchOptions):
                raise ValueError("expected MatchOptions for 'match' WAMP option")

            self["match"] = match.value

        for k, v in kwargs.items():
            self[k] = v


class CallOptions(dict):
    def __init__(self, timeout: int = None, disclose_me: bool | None = None, **kwargs):
        super().__init__()
        if timeout is not None:
            if not isinstance(timeout, int):
                raise ValueError("expected int for 'timeout' WAMP option")

            self["timeout"] = timeout

        if disclose_me is not None:
            if not isinstance(disclose_me, bool):
                raise ValueError("expected bool for 'disclose_me' WAMP option")

            self["disclose_me"] = disclose_me

        for k, v in kwargs.items():
            self[k] = v


class PublishOptions(dict):
    def __init__(
        self,
        acknowledge: bool | None = None,
        exclude_me: bool | None = None,
        disclose_me: bool | None = None,
        exclude: list[int] | None = None,
        eligible: list[int] | None = None,
        exclude_authid: list[int] | None = None,
        eligible_authid: list[int] | None = None,
        exclude_authrole: list[int] | None = None,
        eligible_authrole: list[int] | None = None,
        **kwargs,
    ):
        super().__init__()
        if acknowledge is not None:
            if not isinstance(acknowledge, bool):
                raise ValueError("expected bool for 'acknowledge' WAMP option")

            self["acknowledge"] = acknowledge

        if exclude_me is not None:
            if not isinstance(exclude_me, bool):
                raise ValueError("expected bool for 'exclude_me' WAMP option")

            self["exclude_me"] = exclude_me

        if disclose_me is not None:
            if not isinstance(disclose_me, bool):
                raise ValueError("expected bool for 'disclose_me' WAMP option")

            self["disclose_me"] = disclose_me

        if exclude is not None:
            if not isinstance(exclude, list):
                raise ValueError("expected list for 'exclude' WAMP option")

            self["exclude"] = exclude

        if eligible is not None:
            if not isinstance(eligible, list):
                raise ValueError("expected list for 'eligible' WAMP option")

            self["eligible"] = eligible

        if exclude_authid is not None:
            if not isinstance(exclude_authid, list):
                raise ValueError("expected list for 'exclude_authid' WAMP option")

            self["exclude_authid"] = exclude_authid

        if eligible_authid is not None:
            if not isinstance(eligible_authid, list):
                raise ValueError("expected list for 'eligible_authid' WAMP option")

            self["eligible_authid"] = eligible_authid

        if exclude_authrole is not None:
            if not isinstance(exclude_authrole, list):
                raise ValueError("expected list for 'exclude_authrole' WAMP option")

            self["exclude_authrole"] = exclude_authrole

        if eligible_authrole is not None:
            if not isinstance(eligible_authrole, list):
                raise ValueError("expected list for 'eligible_authrole' WAMP option")

            self["eligible_authrole"] = eligible_authrole

        for k, v in kwargs.items():
            self[k] = v


class Depends:
    def __init__(self, dependency: Callable | Awaitable):
        self._is_async = False
        self._is_async_gen = False

        self.dependency = self._setup(dependency)

    @property
    def is_async(self) -> bool:
        return self._is_async

    @property
    def is_async_gen(self) -> bool:
        return self._is_async_gen

    def _setup(self, func: Callable | Awaitable):
        if inspect.iscoroutinefunction(func):
            self._is_async = True
        elif inspect.isasyncgenfunction(func):
            self._is_async_gen = True
            return contextlib.asynccontextmanager(func)

        return func



import pathlib
import socket

import aiohttp
from aiohttp import web
from wampproto.auth import IServerAuthenticator

from xconn import helpers
from xconn.router import Router
from xconn.acceptor import AIOHttpAcceptor


class Server:
    def __init__(self, router: Router, authenticator: IServerAuthenticator = None):
        self.router = router
        self.authenticator = authenticator

    async def _websocket_handler(self, request):
        protocols = ["wamp.2.json", "wamp.2.cbor", "wamp.2.msgpack"]
        try:
            if helpers._CAPNP_AVAILABLE:
                protocols.append(helpers.CAPNPROTO_SUBPROTOCOL)
        except (ImportError, AttributeError):
            pass

        ws = web.WebSocketResponse(protocols=protocols)
        # upgrade this connection to websocket.
        await ws.prepare(request)

        try:
            acceptor = AIOHttpAcceptor(self.authenticator)
            base_session = await acceptor.accept(ws)
            self.router.attach_client(base_session)
        except Exception:
            await ws.close()

        while not ws.closed:
            msg = await ws.receive()

            if msg.type == aiohttp.WSMsgType.TEXT or msg.type == aiohttp.WSMsgType.BINARY:
                msg = base_session.serializer.deserialize(msg.data)
                await self.router.receive_message(base_session, msg)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                print(f"Error: {msg.exception()}")
            elif msg.type == aiohttp.WSMsgType.CLOSE:
                print("Client disconnected")
                break

        return ws

    async def start(self, host: str, port: int, start_loop: bool = False):
        print(f"starting server on {host}:{port}")
        app = web.Application()
        app.router.add_get("/ws", self._websocket_handler)

        if start_loop:
            web.run_app(app, host=host, port=port)
        else:
            runner = web.AppRunner(app)
            await runner.setup()

            site = aiohttp.web.TCPSite(runner, host=host, port=port)
            await site.start()

    async def start_unix_server(self, socket_path: str) -> None:
        if self._is_unix_socket_alive(socket_path):
            raise RuntimeError(f"Socket at {socket_path} is already in use")

        pathlib.Path(socket_path).unlink(missing_ok=True)

        print(f"Listening on unix://{socket_path}")

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(socket_path)
        sock.listen(128)
        sock.setblocking(False)

        app = web.Application()
        app.router.add_get("/", self._websocket_handler)

        runner = web.AppRunner(app)
        await runner.setup()

        site = web.SockSite(runner, sock)
        await site.start()

    def _is_unix_socket_alive(self, socket_path: str, timeout: float = 1.0) -> bool:
        try:
            with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
                sock.settimeout(timeout)
                sock.connect(socket_path)
            return True
        except (socket.error, ConnectionRefusedError):
            return False



ERROR_RUNTIME_ERROR = "wamp.error.runtime_error"
ERROR_INVALID_ARGUMENT = "wamp.error.invalid_argument"
ERROR_INTERNAL_ERROR = "wamp.error.internal_error"
CLOSE_REALM = "wamp.close.close_realm"
CLOSE_GOODBYE_AND_OUT = "wamp.close.goodbye_and_out"



from asyncio import gather

from wampproto import dealer, broker, messages
from wampproto.types import SessionDetails

from xconn import types, uris


class Realm:
    def __init__(self):
        super().__init__()
        self.dealer = dealer.Dealer()
        self.broker = broker.Broker()

        self.clients: dict[int, types.IAsyncBaseSession] = {}

    def attach_client(self, base: types.IAsyncBaseSession):
        self.clients[base.id] = base

        details = SessionDetails(base.id, base.realm, base.authid, base.authrole)
        self.dealer.add_session(details)
        self.broker.add_session(details)

    def detach_client(self, base: types.IAsyncBaseSession):
        del self.clients[base.id]
        self.broker.remove_session(base.id)
        self.dealer.remove_session(base.id)

    def stop(self):
        """stop will disconnect all clients."""
        pass

    async def receive_message(self, session_id: int, msg: messages.Message):
        match msg.TYPE:
            case (
                messages.Call.TYPE
                | messages.Yield.TYPE
                | messages.Register.TYPE
                | messages.Unregister.TYPE
                | messages.Error.TYPE
            ):
                recipient = self.dealer.receive_message(session_id, msg)
                client = self.clients[recipient.recipient]
                await client.send_message(recipient.message)

            case messages.Publish.TYPE:
                publication = self.broker.receive_publish(session_id, msg)

                if len(publication.recipients) != 0:
                    tasks = []
                    for recipient in publication.recipients:
                        client = self.clients[recipient]
                        tasks.append(client.send_message(publication.event))

                    await gather(*tasks)

                if publication.ack is not None:
                    client = self.clients[publication.ack.recipient]
                    await client.send_message(publication.ack.message)

            case messages.Subscribe.TYPE | messages.Unsubscribe.TYPE:
                recipient = self.broker.receive_message(session_id, msg)
                client = self.clients[recipient.recipient]
                await client.send_message(recipient.message)
            case messages.Goodbye.TYPE:
                self.dealer.remove_session(session_id)
                self.broker.remove_session(session_id)
                try:
                    client = self.clients.pop(session_id)
                except KeyError:
                    return

                goodbye = messages.Goodbye(messages.GoodbyeFields({}, uris.CLOSE_GOODBYE_AND_OUT))
                await client.send_message(goodbye)
                await client.close()



import socket
from typing import Sequence

from aiohttp import web
from wampproto import auth, acceptor, serializers, messages
from websockets import ServerProtocol
from websockets.sync.server import ServerConnection, Subprotocol

from xconn import types, helpers


class WebsocketsAcceptor:
    WSProtocols: Sequence[Subprotocol] = [
        Subprotocol("wamp.2.json"),
        Subprotocol("wamp.2.cbor"),
        Subprotocol("wamp.2.msgpack"),
    ]
    try:
        if helpers._CAPNP_AVAILABLE:
            WSProtocols.append(Subprotocol(helpers.CAPNPROTO_SUBPROTOCOL))
    except (ImportError, AttributeError):
        pass

    def __init__(self, authenticator: auth.IServerAuthenticator = None) -> None:
        self.authenticator = authenticator
        self.subprotocols = WebsocketsAcceptor.WSProtocols

    def accept(self, conn: socket.socket) -> types.BaseSession:
        ws = ServerConnection(conn, ServerProtocol(subprotocols=WebsocketsAcceptor.WSProtocols))
        ws.handshake()

        serializer = helpers.get_serializer(ws.subprotocol)
        a = acceptor.Acceptor(serializer=serializer, authenticator=self.authenticator)

        while True:
            data = ws.recv()
            to_send, is_final = a.receive(data)
            ws.send(to_send)
            if is_final:
                if a.is_aborted():
                    abort: messages.Abort = serializer.deserialize(to_send)
                    raise Exception(abort.reason)

                return types.BaseSession(ws, a.get_session_details(), serializer)


class AIOHttpAcceptor:
    def __init__(self, authenticator: auth.IServerAuthenticator = None) -> None:
        self.authenticator = authenticator

    async def accept(self, ws: web.WebSocketResponse) -> types.AIOHttpBaseSession:
        serializer = helpers.get_serializer(ws.ws_protocol)
        a = acceptor.Acceptor(serializer=serializer, authenticator=self.authenticator)

        if serializer is None or isinstance(serializer, serializers.JSONSerializer):
            send_func = ws.send_str
            receive_func = ws.receive_str
        else:
            send_func = ws.send_bytes
            receive_func = ws.receive_bytes

        while not ws.closed:
            msg = await receive_func()
            to_send, is_final = a.receive(msg)
            await send_func(to_send)
            if is_final:
                if a.is_aborted():
                    abort: messages.Abort = serializer.deserialize(to_send)
                    raise Exception(abort.reason)

                return types.AIOHttpBaseSession(ws, a.get_session_details(), serializer)



import contextlib
import os
import time
import socket
import asyncio
from dataclasses import dataclass
import inspect
import json
import importlib
import sys
from typing import (
    get_type_hints,
    Type,
    Optional,
    Any,
    Callable,
    ContextManager,
    AsyncContextManager,
    Awaitable,
    get_origin,
    Union,
    get_args,
)
from urllib.parse import urlparse

import yaml
from aiohttp import web
from dotenv import load_dotenv
from pydantic import BaseModel, create_model, ValidationError
from wampproto.auth import (
    WAMPCRAAuthenticator,
    TicketAuthenticator,
    CryptoSignAuthenticator,
    AnonymousAuthenticator,
    IClientAuthenticator,
)

from xconn import Router, Server
from xconn.app import App, ExecutionMode
from xconn._client.types import ClientConfig, CommandArgs
from xconn.exception import ApplicationError
from xconn.types import Event, Invocation, Result, Depends, CallDetails, RegisterOptions, SubscribeOptions, EventDetails

MAX_WAIT = 60
INITIAL_WAIT = 1


@dataclass
class BaseMetadata:
    request_model: Type[BaseModel] | None
    request_args: list[str]
    request_kwargs: list[str]

    no_args: bool
    dynamic_model: bool

    dependencies: dict[str, Callable]
    ctx_dependencies: dict[str, ContextManager]
    async_dependencies: dict[str, Awaitable]
    async_ctx_dependencies: dict[str, AsyncContextManager]

    details_field: str | None
    positional_field_name: str | None

    uri: str


@dataclass
class ProcedureMetadata(BaseMetadata):
    response_model: Type[BaseModel] | None
    response_args: list[str]
    response_kwargs: list[str]

    allowed_roles: list[str]

    register_options: dict[str, Any] | RegisterOptions | None
    func: callable


@dataclass
class EventMetadata(BaseMetadata):
    subscribe_options: dict[str, Any] | SubscribeOptions | None
    func: callable


def create_model_from_func(func):
    signature = inspect.signature(func)
    type_hints = get_type_hints(func)
    fields = {}

    for param_name, param in signature.parameters.items():
        annotated_type = type_hints.get(param_name)
        if (
            is_subclass_of_any(annotated_type, CallDetails)
            or is_subclass_of_any(annotated_type, Depends)
            or is_subclass_of_any(annotated_type, EventDetails)
        ):
            continue

        # Handle default values
        if param.default is inspect.Parameter.empty:
            fields[param_name] = (annotated_type, ...)
        else:
            fields[param_name] = (Optional[annotated_type], param.default)

    return create_model(func.__name__.capitalize() + "Model", **fields)


def is_primitive(obj) -> bool:
    return isinstance(obj, (str, int, float, bool, bytes, type(None)))


def is_subclass_of_any(type_, base_class: Any) -> bool:
    origin = get_origin(type_)
    if origin is Union:
        return any(isinstance(arg, type) and issubclass(arg, base_class) for arg in get_args(type_))

    return isinstance(type_, type) and issubclass(type_, base_class)


def _assemble_base_metadata(
    func: callable,
    uri: str,
    incoming_class: type(Invocation) | type(Event),
    details_class: type(CallDetails) | type(EventDetails),
) -> BaseMetadata:
    sig = inspect.signature(func)
    for name, param in sig.parameters.items():
        if param.annotation is inspect._empty:
            raise RuntimeError(f"Missing type hint for parameter: '{name}' in function '{func.__name__}'")

    if issubclass(incoming_class, Event):
        context = "topic"
    elif issubclass(incoming_class, Invocation):
        context = "procedure"
    else:
        raise RuntimeError(f"incoming_class can be with Invocation or Event got='{incoming_class.__name__}'")

    if not issubclass(details_class, CallDetails) and not issubclass(details_class, EventDetails):
        raise RuntimeError(f"details_class can be either CallDetails or EventDetails got='{details_class.__name__}'")

    hints = get_type_hints(func)
    hints.pop("return", None)

    request_model = None
    request_args = []
    request_kwargs = []
    no_args = False
    dynamic_model = False

    dependencies = {}
    ctx_dependencies = {}
    async_dependencies = {}
    async_ctx_dependencies = {}
    for name, param in sig.parameters.items():
        if isinstance(param.default, Depends):
            sig = inspect.signature(param.default.dependency)

            if len(sig.parameters) != 0:
                raise RuntimeError("Dependency functions support no parameters")

            dep: Depends = param.default

            if dep.is_async:
                async_dependencies[name] = param.default.dependency
            elif dep.is_async_gen:
                async_ctx_dependencies[name] = param.default.dependency
            elif inspect.isgeneratorfunction(param.default.dependency):
                ctx_dependencies[name] = contextlib.contextmanager(param.default.dependency)
            elif inspect.isfunction(param.default.dependency) or inspect.ismethod(param.default.dependency):
                dependencies[name] = param.default.dependency

    for key in dependencies.keys():
        del hints[key]

    for key in ctx_dependencies.keys():
        del hints[key]

    for key in async_dependencies.keys():
        del hints[key]

    for key in async_ctx_dependencies.keys():
        del hints[key]

    # check if CallDetails are in the function
    details_field = None
    for name, type_ in hints.items():
        if is_subclass_of_any(type_, details_class):
            if details_field is not None:
                raise RuntimeError(f"Duplicate {details_class.__name__} in function '{func.__name__}'")

            details_field = name

    if details_field is not None:
        del hints[details_field]

    positional_field_name: str | None = None

    has_incoming_class_in_sig = False
    for name, type_ in hints.items():
        if issubclass(type_, BaseModel):
            if len(hints) != 1:
                raise RuntimeError(f"Cannot mix pydantic BaseModel with other types in signature of {context} '{uri}'")

            request_model = type_
            positional_field_name = name
            break

        if is_subclass_of_any(type_, incoming_class):
            if has_incoming_class_in_sig:
                raise RuntimeError(
                    f"Cannot use other types than '{incoming_class.__name__}' as arguments in {context} '{uri}'"
                )

            has_incoming_class_in_sig = True
            positional_field_name = name

    if Invocation in hints.values():
        if len(hints) != 1:
            raise RuntimeError(
                f"Cannot use other types than '{incoming_class.__name__}' as arguments in {context} '{uri}'"
            )
    elif request_model is not None:
        for key, value in request_model.model_fields.items():
            if value.is_required:
                request_args.append(key)
            else:
                request_kwargs.append(key)
    elif len(hints) == 0:
        no_args = True
    else:
        # let's create a dynamic pydantic model based on the procedure signature
        request_model = create_model_from_func(func)
        for key, value in request_model.model_fields.items():
            if value.is_required:
                request_args.append(key)
            else:
                request_kwargs.append(key)

        dynamic_model = True

    return BaseMetadata(
        request_model=request_model,
        request_args=request_args,
        request_kwargs=request_kwargs,
        no_args=no_args,
        dynamic_model=dynamic_model,
        dependencies=dependencies,
        ctx_dependencies=ctx_dependencies,
        async_dependencies=async_dependencies,
        async_ctx_dependencies=async_ctx_dependencies,
        details_field=details_field,
        positional_field_name=positional_field_name,
        uri=uri,
    )


def _validate_procedure_function(func: callable, uri: str) -> ProcedureMetadata:
    meta = _assemble_base_metadata(func, uri, Invocation, CallDetails)

    response_model = getattr(func, "__xconn_response_model__", None)
    response_args = []
    response_kwargs = []
    if response_model is not None:
        for key, value in response_model.model_fields.items():
            if value.is_required:
                response_args.append(key)
            else:
                response_kwargs.append(key)

    allowed_roles = getattr(func, "__xconn_allowed_roles__", [])
    register_options = getattr(func, "__xconn_register_options__", None)

    return ProcedureMetadata(
        request_model=meta.request_model,
        response_model=response_model,
        request_args=meta.request_args,
        response_args=response_args,
        request_kwargs=meta.request_kwargs,
        response_kwargs=response_kwargs,
        no_args=meta.no_args,
        dynamic_model=meta.dynamic_model,
        allowed_roles=allowed_roles,
        dependencies=meta.dependencies,
        ctx_dependencies=meta.ctx_dependencies,
        async_dependencies=meta.async_dependencies,
        async_ctx_dependencies=meta.async_ctx_dependencies,
        details_field=meta.details_field,
        positional_field_name=meta.positional_field_name,
        register_options=register_options,
        uri=meta.uri,
        func=func,
    )


def validate_invocation_parameters(invocation: Invocation, meta: ProcedureMetadata):
    msg = ""
    if invocation.args is not None and len(invocation.args) > len(meta.request_args):
        msg += "expected {} arguments, got {}".format(len(meta.request_args), len(invocation.args))

    if invocation.kwargs is not None and len(invocation.kwargs) > len(meta.request_kwargs):
        if msg != "":
            msg += ", "

        msg += "expected {} keyword arguments, got {}".format(len(meta.request_kwargs), len(invocation.kwargs))

    if msg != "":
        raise ApplicationError("wamp.error.invalid_arguments", msg)


def validate_event_parameters(event: Event, meta: EventMetadata):
    msg = ""
    if event.args is not None and len(event.args) > len(meta.request_args):
        msg += "expected {} arguments, got {}".format(len(meta.request_args), len(event.args))

    if event.kwargs is not None and len(event.kwargs) > len(meta.request_kwargs):
        if msg != "":
            msg += ", "

        msg += "expected {} keyword arguments, got {}".format(len(meta.request_kwargs), len(event.kwargs))

    if msg != "":
        prefix = f"error: function '{meta.func.__name__}' for topic '{meta.uri}' received invalid number of arguments. "
        prefix += msg
        raise ValueError(prefix)


def _handle_result(
    result: Result | tuple | None,
    response_model: type[BaseModel] | None,
    response_positional_args: list[str],
) -> Result:
    if result is None:
        if response_model is not None:
            raise ApplicationError(
                "wamp.error.internal_error", "Procedure returned None, but a response model was provided."
            )

        return Result()

    if isinstance(result, Result):
        if response_model is None:
            return result

        response_args = result.args if result.args is not None else []
        response_kwargs = result.kwargs if result.kwargs is not None else {}

        # If the Result object was returned, we need to be able to map
        # its args to their keys so that the pydantic model can be initialized.
        if len(response_args) != len(response_positional_args):
            raise ApplicationError("wamp.error.internal_error")

        args_with_keys = dict(zip(response_positional_args, response_args))

        # FIXME: catch ValidationError and return ApplicationError
        args = response_model(**args_with_keys, **response_kwargs)
        return Result(args=[args.model_dump()])

    if response_model is None:
        # No response model provided, return the result as-is.
        # We avoid validating the data and shift the responsibility
        # to the serializer.
        return Result(args=[result])

    if is_primitive(result):
        if len(response_positional_args) != 1:
            raise ApplicationError(
                "wamp.error.internal_error",
                f"Procedure returned a single primitive but response model has "
                f"{len(response_positional_args)} positional args.",
            )

        # FIXME: catch ValidationError and return ApplicationError
        args = response_model(**{response_positional_args[0]: result})
        return Result(args=[args.model_dump()])

    # If the result is a tuple, it usually means a python function returned multiple values.
    # Though that might not always be the case, and a function may explicitly return a tuple as well.
    if isinstance(result, tuple):
        if len(result) != len(response_positional_args):
            raise ApplicationError(
                "wamp.error.internal_error",
                f"Procedure returned {len(result)} values but the response model has "
                f"{len(response_positional_args)} args.",
            )

        args_with_keys = dict(zip(response_positional_args, result))

        # FIXME: catch ValidationError and return ApplicationError
        args = response_model(**args_with_keys)
        return Result(args=[args.model_dump()])

    if isinstance(result, list):
        # FIXME: catch ValidationError and return ApplicationError
        return Result(args=[json.loads(response_model.from_orm(item).json()) for item in result])

    # FIXME: catch ValidationError and return ApplicationError
    return Result(args=[json.loads(response_model.from_orm(result).json())])


def _validate_topic_function(func: callable, uri: str):
    meta = _assemble_base_metadata(func, uri, Event, EventDetails)
    options = getattr(func, "__xconn_subscribe_options__", None)

    return EventMetadata(
        request_model=meta.request_model,
        request_args=meta.request_args,
        request_kwargs=meta.request_kwargs,
        no_args=meta.no_args,
        dynamic_model=meta.dynamic_model,
        dependencies=meta.dependencies,
        ctx_dependencies=meta.ctx_dependencies,
        async_dependencies=meta.async_dependencies,
        async_ctx_dependencies=meta.async_ctx_dependencies,
        details_field=meta.details_field,
        positional_field_name=meta.positional_field_name,
        subscribe_options=options,
        uri=meta.uri,
        func=func,
    )


def _sanitize_incoming_data(args: list, kwargs: dict, model_positional_args: list[str]):
    args = args if args is not None else []
    kwargs = kwargs if kwargs is not None else {}

    args_with_keys = dict(zip(model_positional_args, args))
    args_with_keys.update(kwargs)

    return args_with_keys


def collect_docs(uri: str, func: callable, type_: str):
    meta = _validate_procedure_function(func, uri)

    data = {"uri": uri, "type": type_}
    if meta.request_model is not None:
        data["in_model"] = meta.request_model.model_json_schema()

    if meta.response_model is not None:
        data["out_model"] = meta.response_model.model_json_schema()

    return data


def create_app(docs: list[dict], endpoint: str):
    app = web.Application()

    async def serve_schema(_):
        return web.json_response(docs)

    app.router.add_get(endpoint, serve_schema)
    return app


def select_authenticator(config: ClientConfig) -> IClientAuthenticator:
    if config.authmethod == "cryptosign" or config.authmethod == "wampcra" or config.authmethod == "ticket":
        if config.authmethod == "wampcra":
            auth = WAMPCRAAuthenticator(config.authid, config.secret)
        elif config.authmethod == "ticket":
            auth = TicketAuthenticator(config.authid, config.ticket)
        else:
            auth = CryptoSignAuthenticator(config.authid, config.private_key)
    else:
        auth = AnonymousAuthenticator(authid=config.authid)

    return auth


async def start_server_async(config: ClientConfig):
    r = Router()
    r.add_realm(config.realm)
    server = Server(r)
    url_parsed = urlparse(config.url)
    await server.start(url_parsed.hostname, url_parsed.port)


def start_server_sync(config: ClientConfig):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(start_server_async(config))
    loop.run_forever()


def is_non_empty(value):
    return value is not None and value != ""


def validate_auth_inputs(config: ClientConfig) -> None:
    if is_non_empty(config.private_key) and is_non_empty(config.ticket):
        raise ValueError("provide only one of private key, ticket or secret")
    elif is_non_empty(config.ticket) and is_non_empty(config.secret):
        raise ValueError("provide only one of private key, ticket or secret")
    elif is_non_empty(config.private_key) and is_non_empty(config.secret):
        raise ValueError("provide only one of private key, ticket or secret")

    if config.authid is None and any(
        is_non_empty(value) for value in (config.private_key, config.ticket, config.secret)
    ):
        raise ValueError("authid is required with any of private key, ticket or secret")


def select_authmethod(config: ClientConfig) -> str:
    if is_non_empty(config.private_key) and not config.ticket and not config.secret:
        return CryptoSignAuthenticator.TYPE
    elif is_non_empty(config.ticket) and not config.private_key and not config.secret:
        return TicketAuthenticator.TYPE
    elif is_non_empty(config.secret) and not config.private_key and not config.ticket:
        return WAMPCRAAuthenticator.TYPE

    return AnonymousAuthenticator.TYPE


def wait_for_server(host: str, port: int, timeout: float):
    start_time = time.time()
    while time.time() - start_time < timeout:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.connect((host, port))
                return True
            except (ConnectionRefusedError, socket.timeout):
                time.sleep(0.2)

    raise TimeoutError(f"Server did not start listening on {host}:{port} within {timeout} seconds.")


def handle_model_validation(model, **kwargs):
    try:
        return model(**kwargs)
    except ValidationError as e:
        raise ApplicationError("wamp.error.invalid_argument", e.json())


def ensure_caller_allowed(call_details: dict[str, Any], allowed_roles: list[str]):
    if len(allowed_roles) == 0:
        return

    role = call_details.get("caller_authrole", None)
    if role is None:
        msg = (
            "Router did not send call details hence the authrole of the caller cannot be determined."
            f" The caller must have one of following roles '{allowed_roles}'"
        )
        raise ApplicationError("wamp.error.internal_error", msg)

    if role not in allowed_roles:
        msg = f"The caller must have one of following roles '{allowed_roles}' got={role}"
        raise ApplicationError("wamp.error.not_authorized", msg)


def assemble_call_details(uri: str, meta: ProcedureMetadata, invocation: Invocation):
    details = {}
    if meta.details_field is not None:
        if not invocation.details:
            msg = f"Endpoint for procedure {uri} expects CallDetails but router did not send them"
            raise ApplicationError("wamp.error.internal_error", msg)

        details[meta.details_field] = CallDetails(invocation.details)

    return details


def assemble_event_details(uri: str, meta: EventMetadata, event: Event):
    details = {}
    if meta.details_field is not None:
        if not event.details:
            msg = f"Endpoint for topic {uri} expects EventDetails but router did not send them"
            raise ApplicationError("wamp.error.internal_error", msg)

        details[meta.details_field] = CallDetails(event.details)

    return details


def import_app(app: str, directory: str = ".") -> App:
    split = app.split(":")
    if len(split) != 2:
        raise RuntimeError("invalid app argument, must be of format: module:instance")

    sys.path.append(directory)
    module = importlib.import_module(split[0])
    app: App = getattr(module, split[1])
    if not isinstance(app, App):
        raise RuntimeError(f"app instance is of unknown type {type(app)}")

    return app


def connect(app: str, config: ClientConfig, start_router: bool = False, directory: str = "."):
    imported_app = import_app(app, directory)
    if imported_app.execution_mode == ExecutionMode.ASYNC:
        from xconn._client.async_ import _connect_async
        from xconn.utils import run

        run(_connect_async(imported_app, config, start_router=start_router))
    elif imported_app.execution_mode == ExecutionMode.SYNC:
        from xconn._client.sync import _connect_sync

        _connect_sync(imported_app, config, start_router=start_router)
    else:
        raise RuntimeError(f"execution mode {imported_app.execution_mode} not supported yet")


def load_config_from_env(args: CommandArgs) -> ClientConfig:
    if args.url is not None and args.url != "":
        url = args.url
    else:
        url = os.environ.get("XCORN_URL", None)
        if url is None or url == "":
            print("XCORN_URL missing in environment variable")
            exit(1)

    if args.realm is not None and args.realm != "":
        realm = args.realm
    else:
        realm = os.environ.get("XCORN_REALM", None)
        if realm is None or realm == "":
            print("XCORN_REALM missing in environment variable")
            exit(1)

    return ClientConfig(
        url=url,
        realm=realm,
        authid=os.environ.get("XCORN_AUTHID", args.authid),
        secret=os.environ.get("XCORN_SECRET", args.secret),
        ticket=os.environ.get("XCORN_TICKET", args.ticket),
        private_key=os.environ.get("XCORN_PRIVATE_KEY", args.private_key),
    )


def load_config_from_dot_env(args: CommandArgs):
    if args.config_file is not None and args.config_file != "":
        load_dotenv(args.config_file)
    elif os.path.exists(os.path.join(os.getcwd(), ".env")):
        load_dotenv()
    else:
        print(".env file doesn't exist. You may use --config-file flag to specify a env file.")
        exit(1)

    return load_config_from_env(args)


def load_config_from_yaml(args: CommandArgs) -> ClientConfig:
    with open(args.config_file) as f:
        config_raw = yaml.safe_load(f)

    for name, value in args.model_dump().items():
        if value is not None:
            config_raw[name] = value

    return ClientConfig(**config_raw)


def load_config_from_file(args: CommandArgs) -> ClientConfig:
    if args.config_file is not None:
        if not os.path.exists(args.config_file):
            raise FileNotFoundError(args.config_file)

    try:
        return load_config_from_yaml(args)
    except Exception:
        return load_config_from_dot_env(args)


def update_config_from_cli(config: ClientConfig, command_args: CommandArgs) -> ClientConfig:
    command_args_dict = command_args.model_dump(exclude_none=True)

    for key, value in command_args_dict.items():
        if hasattr(config, key):
            setattr(config, key, value)

    return config



import asyncio
from typing import Coroutine
import signal


def run(main: Coroutine):
    async def _run():
        loop = asyncio.get_running_loop()
        stop_event = asyncio.Event()

        def handle_signal(sig):
            print(f"\nReceived signal {sig.name}, stopping...")
            stop_event.set()

        for sig in (signal.SIGINT, signal.SIGTERM):
            loop.add_signal_handler(sig, handle_signal, sig)

        await main
        await stop_event.wait()

    asyncio.run(_run())
